/*
** Module   : Frame.java
** Abstract : Frame implementation for CHUI
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 EVL 20050324   @20506 Creation based on CHARVA JFrame class.
**                           Necessary modification to get compiled.
** 002 EVL 20050405   @20655 Changing class name to the Frame.
** 003 EVL 20050406   @20656 Adding scrollable CHUI Container.
** 004 EVL 20050412   @20685 Removed the redundant drawing code. Now
**                           The frame border is drawing inside the
**                           ScrollContainer
** 005 EVL 20050413   @20831 Adding support for frame families.
** 006 EVL 20050424   @20860 Adding scrollable flag.
** 007 EVL 20050503   @21071 Adding centering flag and handling this case.
** 008 EVL 20050503   @21072 Moving one line border inside the ScrollPane.
** 009 EVL 20050505   @21114 Adding helper functions to implement the
**                           Progress specific methods.
** 010 EVL 20050506   @21130 Isolate all Progress specific code in helper
**                           class FrameHelper.
** 011 EVL 20050509   @21170 Adding parenting to the main Window widget.
** 012 EVL 20050511   @21171 Adding focus request to the next component.
** 013 EVL 20050514   @21175 Adding the ability to query frame centering
**                           flag.
** 014 EVL 20050516   @21182 Changing the Z-ordering handling approach.
** 015 EVL 20050516   @21183 Adding the Progress specific methods.
** 016 EVL 20050526   @21327 Changing the inner variable names to sync
**                           with the rest of the project.
** 017 EVL 20050527   @21328 Adding handling of the Progress frame widget
**                           attributes and preparing to javadoc
**                           processing.
** 018 EVL 20050530   @21338 Changing method name for working with parent
**                           attribute.
** 019 EVL 20050601   @21371 Adding the method to get the FrameHelper
**                           instance.
** 020 EVL 20050607   @21435 Syntax for comments checking.
** 021 EVL 20050607   @21438 Removing the visible member. It is completely
**                           duplicated in CHARVA's Component superclass
**                           so we do not need to double this work. Also
**                           the methods to work with the HIDDEN attribute
**                           are modified to reflect this change.
** 022 EVL 20050608   @21447 Adding the method to put the widget in a
**                           frame without explicitly defining the
**                           coordinates. Also the method addSkip() was
**                           added to put the number of empty lines into 
**                           the frame.
** 023 EVL 20050615   @21512 Adoption the attribute to the CHARVA model.
**                           The attributes TAB-POSITION, ROW, COLUMN,
**                           FRAME, VIRTUAL-WIDTH/HEIGHT-CHARS are
**                           touched.
** 024 EVL 20050620   @21548 Adding set(do)Layout methods.
** 025 EVL 20050628   @21560 Adding the method addSpace(n) to be able to
**                           add the empty column gap to the frame.
** 026 EVL 20050822   @22199 Unifying containers by adding P2JContainer
**                           interface implementation.
** 027 SIY 20050831   @22503 Refactoring (see @@22394-22409) 
** 028 SIY 20050907   @22636 Fixed use of @inheritDoc tag in javadoc
** 029 EVL 20050915   @22730 Adding method to remove the key listener for
**                           the frame class.
** 030 SIY 20050920   @22827 Added support for SENSITIVE attribute.
** 031 SIY 20060128   @24118 Added support code for the frame manager.
** 032 SIY 20060201   @24206 Removed reference to Window. Javadoc
**                           cleanups.
** 033 SIY 20060202   @24236 setUnused() automatically removes frame if
**                           necessary.
** 034 SIY 20060209   @24468 Updated to accommodate changes in xxxConfig
**                           classes. Minor cleanups.
** 035 SIY 20060213   @24575 Added support for DialogBox mode; added 
**                           helper methods.
** 036 SIY 20060222   @24638 Implemented CHOOSE processing.
** 037 NVS 20060227   @24787 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 038 SIY 20060315   @25066 Additional flags used for frame hiding.
** 039 SIY 20060316   @25130 Fixes and updates caused by header support.
** 040 SIY 20060323   @25219 Changes caused by new approach in down frame.
** 041 SIY 20060411   @25596 Down frame size calculation fix.
** 042 SIY 20060424   @25682 Encapsulated access to down field.
** 043 GES 20060503   @25964 Fixed typo in method name.
** 044 SIY 20060510   @26065 More careful tracking of hadFocus flag.
** 045 SIY 20060525   @26538 Fixed handling of key events in CHOOSE.
** 046 SIY 20060525   @26546 Moved part of common doLayout() code into
**                           separate adjustDown() method and added other
**                           convenience methods.
** 047 NVS 20060526   @26582 Added "realized" state flag with its getter
**                           and one way setter.
** 048 SIY 20060531   @26909 Fixed handling of "realized" state flag.
** 049 SIY 20060601   @26910 New approach in pause handling.
** 050 SIY 20060610   @27117 Fixed tracking of the canHide flag.
** 051 SIY 20060613   @27229 Fixed relayout() method.
** 052 SIY 20060615   @27238 More fixes in relayout() and related methods.
** 053 SIY 20060616   @27343 Now doLayout() is safe for repeated calls and
**                           takes into account that order of widgets can
**                           be changed during normal work (widget tab 
**                           order depends on enable() calls).
** 054 SIY 20060616   @27366 Fixed regression which exposed as incorrect 
**                           handling of frame properties (labels and
**                           other flags).
** 055 SIY 20060616   @27398 Cleaned up experimental and debugging code. 
** 056 SIY 20060617   @27420 Down calculation fix. 
** 057 SIY 20060622   @27535 Added toString() method which is more useful
**                           and informative than inherited one.
** 058 GES 20060622   @27538 Fix for array out of bounds failure.
** 059 SIY 20060627   @27661 Fix for layout handling when title is longer
**                           than calculated frame width. Added support 
**                           for dynamic down.
** 060 GES 20060630   @27736 Code cleanup and changed some signatures.
**                           Safety code was added and new clearing of
**                           the keys list for cursor movement keys was
**                           added.
** 061 SIY 20060703   @27749 Minor change to prevent screwing up saved 
**                           value.
** 062 NVS 20060630   @27771 Reviewed requestFocus() calls made from here
**                           and found them safe, having no Progress
**                           semantics.
** 063 SIY 20060705   @27835 Initial value of the choose in down frame is
**                           set to current frame line.
** 064 SIY 20060717   @28093 Take into account rounding for down frames 
**                           with iteration height more than 1.
** 065 SIY 20060721   @28142 Fixed getFrameValue(), many other minor 
**                           changes and fixes. 
** 066 NVS 20060728   @28210 TempContainer class implements replace()
**                           method that updates the non-body list for down
**                           frames.
** 067 SIY 20060726   @28230 Added support for frame resetting on hide,
**                           fixed getFrameValue(), other changes related
**                           to new frame hiding algorithm.
** 068 GES 20060731   @28243 Added getChooseData(). 
** 069 SIY 20060731   @28253 Frame cleanup changes.  
** 070 NVS 20060731   @28260 choose() avoids extra frame repainting and
**                           also draws the selection directly.
** 071 NVS 20060801   @28291 leaveChoose() uses direct drawing to restore
**                           the down frame state.
** 072 SIY 20060802   @28336 Protect against repeated calls to reset().
** 073 GES 20060803   @28407 Minor changes in border chars interface plus
**                           some cleanup.
** 074 GES 20060808   @28489 Minor changes to match new signature.
** 075 SIY 20060822   @28729 More careful handling of getFrameValue() calls 
**                           not related to CHOOSE processing.
** 076 NVS 20060823   @28751 doLayout() method calls detectTrailingSkips()
**                           method of FrameHelper.
** 077 GES 20060824   @28837 Removed hard coding of dialog boxes to fail
**                           isFocusTraversable().
** 078 SIY 20060824   @28834 Simplified frame size setting logic and added
**                           helper method for FrameHelper.
** 079 SIY 20060824   @28851 Removed getP2JType() method.
** 080 SIY 20060825   @28891 Encapsulated FrameConfig fields.
** 081 SIY 20060828   @28924 Temporary container should have parent set
**                           properly.
** 082 SIY 20060831   @29206 Minor changes in comments.
** 083 SIY 20060904   @29211 Always use the TempContainer for size/layout.
**                           This resolves regressions in frame layout 
**                           handling caused by the fact that down body 
**                           now is the part of the content pane.
** 084 GES 20060921   @29837 Dialogs without focusable widgets cause a
**                           failure during nextFocus() processing if
**                           the hasFocus() method is forced to return
**                           true.  This limitation was removed and seems
**                           to have been a vestige of the past (exactly
**                           like the same code that was removed from
**                           isFocusTraversable() previously).
** 085 NVS 20060926   @29930 Added allClear flag and getter/setter for it.
**                           This flag helps in tracking of the initial or
**                           all cleared state of the frame. While in this
**                           state, conditional UPs and DOWNs are ignored.
** 086 NVS 20060927   @29971 Fixed CHOOSE highlight drawing that sometimes
**                           would fail due to an arbitrarily set clipping
**                           rectangle. Optimized highlight drawing for
**                           successive CHOOSE uses.
** 087 NVS 20061005   @30163 Extended getFrameValue() method to select the
**                           target widget.
** 088 SIY 20061010   @30270 Place frame in setVisible() if necessary.
** 089 SIY 20061012   @30371 Removed debugging code.
** 090 NVS 20061020   @30584 Added a variable and methods to manage the
**                           frame's overlay rectangle.
** 091 GES 20061027   @30779 Removed "alive" member/usage and cleaned up
**                           javadoc.
** 092 NVS 20061103   @30951 Put a TO-DO note in adjustDown() method.
** 093 SIY 20061117   @31290 Updated for support of dynamic row/column. 
** 094 SIY 20061128   @31431 Dynamically calculated DOWN value must be 
**                           greater than 0. 
** 095 NVS 20061205   @31563 Fixed getFrameValue() in case a SelectionList
**                           Body is in focus. The owning SelectionList
**                           has to be queried.
** 096 SIY 20070116   @31842 Remove frame from the list invisible, out of 
**                           scope frames.
** 097 SIY 20070125   @31971 Remove frame directly from the default window.
** 098 SIY 20070214   @32131 Include delimiter into list of widgets
**                           maintained by TempContainer.
** 099 SIY 20070306   @32281 Removing frame now is done in ThinClient.
**                           Cleaned up unused 'level' field. 
** 100 SIY 20070312   @32394 Added active scope tracking.
** 101 SIY 20070412   @32958 Pass through PAGE-UP/-DOWN.
** 102 SIY 20070413   @32968 Do not set "internal GO" flag for PG-UP/-DOWN.
** 103 SIY 20070420   @33157 Fixed initial comparison in setVisible() which
**                           prevented invocation of placeFrame() for 
**                           frames which have only Y coordinate undefined. 
** 104 NVS 20070426   @33253 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 105 SIY 20070501   @33338 Removed redundant widget adding code and 
**                           reference to P2JFrame. 
** 106 SIY 20070516   @33749 Added support for delayed focus activation
**                           component ID handling. Fixed javadoc. Fixes
**                           in CHOOSE handling. Added scope tracking flag.
** 107 SIY 20070703   @34390 Fixed isFocusTraversable for case when choose
**                           is active.
** 108 SIY 20070712   @34464 Fixed CLEAR/CLEAR ALL for down frames.
** 109 SIY 20070731   @34738 Fixed pausing of redirected frames.
** 110 SIY 20070814   @34846 Track frame redirection.
** 111 NVS 20070815   @34849 Added isRedirected() method.
** 112 NVS 20070828   @34922 Frame redirection also tracks the last used
**                           stream by its ID. Added getStreamId() method
**                           to query this value
** 113 NVS 20070830   @34975 Added support for the new streamed flag:
**                           markAsStreamed() and isStreamed() methods.
** 114 NVS 20070925   @35210 Added debugging aid method view(). Useful in
**                           the debugging session to print the contents
**                           of the terminal buffers with "invoke view()".
** 115 SIY 20071030   @35640 Fixed down frame height calculation in certain
**                           cases. 
**                           Ref: #35639
** 116 EVL 20071031   @35652 Fix for menu function browser incorrect key
**                           handling after pressing quick navigation keys.
**                           See @35653 for more details.
** 117 SIY 20071031   @35657 Fix for redundant pause.
**                           Ref: #35656
** 118 SIY 20071102   @35685 In redirected mode draw even frames with zero 
**                           width, because it may contain SKIP(n).
** 119 SIY 20071105   @35713 More support methods for correct SKIP handling 
**                           in redirected mode.
** 120 SIY 20071108   @35757 Implemented Progress' quirk in frame height 
**                           calculation algorithm.
**                           Ref: #35755
** 121 EVL 20071116   @35827 Fix for TempText overlapping by choose
**                           processing. The frame border redraws itself
**                           every time the choose scrolls or initially
**                           being shown. So every TempText is placed on
**                           frame border must be moved on the top.
** 122 SVL 20071206   @36228 restoreInitialTabOrder() added
** 123 GES 20071213   @36328 Fixes to choose processing to handle wrapping
**                           and cursor movement deviations. In row mode
**                           choose now sets the down frame's cursor to
**                           ensure that the frame-line is correct.
** 124 SIY 20071214   @36361 Track color changes.
**                           Ref: #36348
** 125 CA  20071214   @36334 Fixed the "display stream s skip(2)" - 
**                           the width of the computed rectangle should be 
**                           1, not 0, by overriding the getSize() method.
**                           Ref: #36368
** 126 SIY 20071214   @36391 Conditional frame hiding reset saved focus ID.
**                           Ref: #36390
** 127 EVL 20071220   @36506 Fix for nextField/prevField algorithms.
** 128 NVS 20071218   @36461 reset() method should reset 1 down frames as
**                           well as down frames.
** 129 NVS 20071229   @36611 setRedirected() method hides the frame when
**                           stream switching has occurred and the new
**                           stream has no frame headers yet.
** 130 CA  20080109   @36738 Back out H125. The fix for #36368 (display
**                           with only skip statement) should have been in 
**                           doLayout(). 
** 131 NVS 20080114   @36807 Implemented fixed height calculation per SIY
**                           suggestions. Now it's maintained for all
**                           frames.
** 132 CA  20080124   @36897 Added fields delayedRepaint, scopeNesting and
**                           accessors. Ref: #36895
** 133 SIY 20080129   @36930 Moved hiding of TempText into setVisible() to
**                           assure that temp text will be hidden 
**                           regardless from the way in which frame is get
**                           hidden.
**                           Ref: #36929
** 134 SIY 20080214   @37109 Frame width calculation-related cleanups 
**                           and refinements, convenience methods added.
**                           Ref: #37107
** 135 SIY 20080303   @37290 Replaced calls to FrameHelper.toTop() and 
**                           FrameHelper.toBottom() with appropriate 
**                           Container methods. 
** 136 SIY 20080320   @37603 Fixed handling of SelectionList.
** 137 NVS 20080320   @37620 choose should honor any outstanding
**                           prompt-for request.
**                           Added supporting code for maintaining a list
**                           of displayed frames.
** 138 NVS 20080328   @37725 Partial back out of #137. List of displayed
**                           frames is no longer maintained.
** 139 NVS 20080401   @37776 Added isFocusable() method implementation.
** 140 SIY 20080403   @37799 Do not pass down event targeted to frame
**                           itself.
** 141 GES 20080421   @38061 I18N charset support.
** 142 NVS 20080423   @38102 Major reimplementation of CHOOSE in field
**                           mode to support triggers and regular focus
**                           change.
** 143 NVS 20080428   @38129 Row mode CHOOSE needed a different initial
**                           redrawing approach.
** 144 SIY 20080430   @38156 Changed initial state of allClear flag. It 
**                           was working as a protection against DOWN 
**                           applied to not viewed frames, but this is not
**                           necessary anymore (Ref: #38234,38137).
** 145 NVS 20080502   @38199 Minor change to unify use of tryFocusChange()
** 146 SIY 20080505   @38203 When frame goes out of scope NO-HIDE flag is 
**                           no longer maintained (updated by NVS).
**                           Ref: #38202
** 147 SIY 20080516   @38290 Suppress frame redraw if only widget color is
**                           changed.
** 148 SIY 20080528   @38427 Postpone changing visibility and placing frame
**                           until widget is completely reconstructed.
**                           Latent issue exposed by #38289,38290. 
** 149 CA  20080529   @38433 Fixed bug introduced by #38203 - when the
**                           frame re-opens the scope, the no-hide status
**                           must be restored to the value set at config
**                           time. Ref: #38431
** 150 SIY 20080602   @38468 Save DOWN value after instantiation of
**                           FrameHelper (which may adjust DOWN in some
**                           cases). 
**                           Ref: #36974 
** 151 SIY 20080605   @38572 A number of small changes required to avoid
**                           uncontrolled drawing in pushScreenDefinition.
** 152 NVS 20080605   @38576 Removed variable and methods supporting
**                           frame overlay rectangle tracking feature.
**                           Since implementation of invalidation
**                           rectangle tracking for paint events, it's no
**                           longer needed and was causing excessive
**                           frame repaints under certain conditions.
** 153 SIY 20080630   @38988 Added separate protection flag which allows
**                           protect frame from removal when its visibility
**                           is set to false.
**                           Ref: #38987
** 154 SIY 20080715   @39112 Renamed parameter to reflect its real meaning.
** 155 SIY 20080731   @39259 Added forced pause flag.
** 156 SIY 20080808   @39344 Frame does not need forced refresh if it is 
**                           get hidden.
** 157 NVS 20080812   @39369 Introduced a new frame state flag telling
**                           whether the frame contains editors and a
**                           getter for this state flag, hasEditors()
** 158 NVS 20080828   @39617 Added isClipping() method which tells that
**                           this container allows clipping of Paint
**                           events.
** 159 CA  20080905   @39745 Changed shortName() from private to package
**                           private.
** 160 SIY 20080922   @39899 Added 'flushed' flag.
** 161 SIY 20081003   @40004 More strict filtering of drawing requests for
**                           interactive mode.
** 162 SIY 20081003   @40046 Added flag which tracks last frame placement
**                           mode - interactive vs redirected.
** 163 CA  20080930   @39984 Fixed frame re-size via WIDTH/HEIGHT-CHARS
**                           attributes.
** 164 SIY 20081020   @40170 Method which places frame in proper Z-order
**                           is moved from ThinClient to Frame.
** 165 SIY 20081105   @40369 Added flag which tracks 'focus attractor'
**                           state. See #40367 for more details.
** 166 CA  20081118   @40463 Do not compute the dynamic-down value if the
**                           frame was not yet placed.
** 167 CA  20081114   @40444 Changes due to new clipping implementation.
** 168 SVL 20081118   @40494 start/endSizeChange() calls added to
**                           doLayout(boolean) and setSize(int,int).
** 169 CA  20081127   @40703 Added 'canHideOthers' flag, indicating that
**                           this frame can hide (or not) other frames.
** 170 NVS 20081205   @40811 Standardized the use of getAndResetBitmap().
** 171 CA  20081219   @40951 For DOWN and fixed-size frames, their DOWN
**                           value is computed from the frame height.
** 172 SIY 20081218   @40955 Turn off placement quirk mode if frame with
**                           appropriate row is get hidden.
** 173 SIY 20081224   @40992 Reworked #172.
** 174 SVL 20081230   @41000 Original frame height is preserved into 
**                           adjustDown().
** 175 CA  20090106   @41021 Track frame hiding via explicit HIDE calls.
** 176 CA  20090113   @41123 Track frame visibility change, from a not 
**                           visible to a visible state.
** 177 NVS 20090209   @41280 Added calculateInnerRect() helper. Changed
**                           the way PUT SCREEN output gets cleaned when
**                           the frame hides. Now it's done through a call
**                           to the refresh() method of the double
**                           buffering output driver.
** 178 NVS 20090212   @41300 Renamed calculateInnerRect() helper method
**                           into calculateFrameRect() and extended its
**                           function to conditionally exclude the box.
**                           setVisible() forces frame rectangle refresh
**                           through double buffering output driver every
**                           time the frame changes visibility. Refresh
**                           includes the frame border for view, not for
**                           hide. Syncing is not performed instantly and
**                           this is not an oversight. setVisible() is
**                           called from methods that sync themselves.
** 179 NVS 20090213   @41314 Dead code removal.
** 180 NVS 20090225   @41362 Frame hiding refreshes the screen through the
**                           double buffering driver so that the PUT
**                           SCREENs that were landed on the right side of
**                           the frame on the unoccupied screen lines are
**                           wiped out (frame_hiding_clears_put_screen*.p)
** 181 SIY 20090311   @41507 Disable forced pause if needPause is set to
**                           false.
** 182 SIY 20090319   @41643 Implemented proper PUT SCEEN removal on frame
**                           hiding.
** 183 SIY 20090324   @41670 Implemented support for partial refreshing on
**                           initial frame display.
** 184 SIY 20090409   @41736 Fixed handling of fixed size frame in cases 
**                           when layout requires more space than specified
**                           in frame definition.  
** 185 CA  20090413   @41779 On init, ENABLE flag must be ON.
** 186 SIY 20090422   @41841 Added canWipeOthers flag.
** 187 CA  20090422   @41838 Added handling of NEXT-FRAME and PREV-FRAME 
**                           events.
** 188 SIY 20090428   @42015 Support for saving/restoring frame state.
** 189 SIY 20090501   @42088 Use notifications about stream end of life when
**                           restoring frame state. 
** 190 SIY 20090604   @42557 Do not change frame visibility on color change 
**                           requests.
** 191 SIY 20090615   @42694 Slightly changed frame refreshing on visibility
**                           change algorithm in order to make it less 
**                           aggressive.
** 192 SIY 20090701   @43018 Added helper methods for format checking.
** 193 SIY 20090707   @43068 Fixed refreshing of the frame border.
** 194 SIY 20090730   @43459 hasTitle() now is package private.
** 195 SIY 20090821   @43714 Fixed hasEnabled() method. It should not take into
**                           account labels, skips, etc.
** 196 SIY 20090818   @43688 Added getClipRect() method.
** 197 SIY 20090824   @43730 hasEnabled() split into two separate methods with
**                           different semantics.
** 198 SIY 20090828   @43791 Added 'force wipe screen' flag.
** 199 SIY 20090902   @43825 Added 'focus reset' flag. 
** 200 SIY 20091015   @44135 Refactoring which is necessary to expose common
**                           code as containsFrames() method.
** 201 SIY 20091020   @44175 Frames which were not displayed (because they are
**                           empty) should not trigger pause.
** 202 SIY 20100115   @44528 Use dedicated algorithm for refreshing of 
**                           horizontal frame lines.
** 203 SIY 20100119   @44532 Narrowed conditions for use of new algorithm of
**                           refreshing horizontal frame lines.
** 204 SIY 20100127   @44551 Adjusted refreshing code for proper handling of 
**                           some edge cases.
** 205 SIY 20100129   @44563 Fixed refreshing on different types of terminal.
** 206 SIY 20100316   @44750 Yet another per-frame flag.
** 207 SIY 20100619          Switching to own implementation of character UI.
** 208 SIY 20100804          Changes in imports, some refactorings.
** 209 SIY 20100822          Refactored AttributeContainer support.
** 210 SIY 20100831          Refactored FrameHelper and responsibilities
**                           split in more correct way.
** 211 SIY 20100915          Deep refactoring.
** 212 EVL 20100920          Fixing the widgets underline for empty widget case.
**                           The change consist of:
**                           1. Merging the underline flags in fillDownBody()
**                              when P2J already planned to underline something
**                           2. The method draw() has been extended in the
**                              underline clause. Now P2J relocates and draws
**                              the widget itself after rendering underline.
**                           These changes are touched the report generation
**                           mode and only for empty decimal FillIn instances
**                           at row number 0.
** 213 SIY 20101015          Changes related to split of Screen into two
**                           separate classes and related refactorings.
** 214 SIY 20101122          Switching from KeyEvent to KeyInput.
** 215 SIY 20110112          Changes related to new color handling scheme.
** 216 SIY 20111101          Updated imports.
** 217 SIY 20111202          Refactored and moved from CHUI package.
** 218 SIY 20111221          Switched to widget factory.
** 219 CA  20130927          copyAffectedWidgets must return an empty array if no widget is found.
*/
 
package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;


/**
 * A <code>Frame</code> is a top-level window with a title and a border. 
 */
public abstract class Frame
extends BorderedPanel
implements PageBreakListener,
           CloseListener
{
   /** Track if frame was viewed inside scope. */
   private boolean viewed = false;
    
   /** Track frame life time. */
   private boolean live = true;

   /** <code>true</code> if the frame can be hidden. */
   private boolean canHide = false;

   /** <code>true</code> if the frame can hide other frames. */
   private boolean canHideOthers = false;

   /** <code>true</code> if the frame can wipe screen when hiding other frames. */
   private boolean canWipeScreen = false;

   /** <code>true</code> if the frame must wipe screen on next view call. */
   private boolean forceWipeScreen = false;
   
   /** Automatic placing is performed for the frame. */
   private boolean autoplacing = true;

   /** Saved DOWN attribute. */
   private int savedDown = -1;
   
   /** Do automatic layout calculation. */
   private boolean autoLayout = false;

   /** Set to true when frame requires pause before hiding. */
   protected boolean needPause = true;
   
   /** Saved original widget data. */
   private WidgetItem[] savedWidgets = null;

   /** Saved frame dimension. */
   private Dimension savedDim = null;

   /** Saved frame pane dimension. */
   private Dimension savedPaneDim = null;

   /** Track dynamic down. */
   private boolean dynamicDown = false;

   /** New value for the dynamic down frame. */
   private int pendingDown = -1;

   /** "frame streamed" flag */
   private boolean streamed = false;
 
   /**
    * Pending reset. When <code>true</code>, frame widgets need to be reset
    * when frame is about to be displayed.
    */
   private boolean resetPending = false;
   
   /** Count the scope nesting level. */
   private int scopeNesting = 0;

   /**
    * Pending repaint. When <code>true</code>, frames will be repainted on 
    * the next display/view statement.
    */
   private boolean delayedRepaint = false;

   /** Reflects paintable state of frame */
   private boolean paintable = true;

   /** Reflects the initial/all clear state of frame */
   private boolean allClear = false;
   
   /** Dynamically calculated column. */
   private int dynamicCol = BaseConfig.INV_COORD;

   /** Dynamically calculated row. */
   private int dynamicRow = BaseConfig.INV_COORD;
   
   /** Active scope flag. */
   private boolean active = false;

   /** Delayed focus activation component ID. */
   private int focusId = -1;

   /** Frame in scope flag. */
   private boolean inScope = false;

   /** Set to <code>true</code> if frame is redirected to some stream. */
   protected boolean redirected = false;
   
   /** the last used stream ID in redirected mode */
   private int streamId = -1;

   /** Track frames with fixed (as opposed to dynamically assigned) row. */
   private boolean isFixedRow = false;

   /** Track start/end of config update. */
   private boolean configUpdateActive = false;

   /** Track color changes. */
   private boolean colorChanged = false;

   /** Postponed visibility change flag.*/
   private boolean postponedVisible = false;

   /** Postponed placing frame on screen. */
   private boolean postponedPlace = false;

   /** Protect frame from removal on hide.*/
   private boolean protect = false;

   /** Track forced pause state. */
   private boolean forcePause = false;

   /** Frame contains editors */
   private boolean withEditors = false;

   /** Track flushed frame state. */
   private boolean flushed = false;
   
   /** <code>true</code> if placement was done for redirected mode. */
   private boolean placedForRedirected = false;
 
   /** 'focus attractor' flag */
   private boolean focusAttractor = false;
   
   /** Track frame visibility change, from not-visible to visible. */
   private boolean visibilityChanged = false;

   /** Flag tracking frame hiding via explicit HIDE calls. */
   private boolean explicitHide = false;

   /** Bound streams. */
   private Set<Integer> streams = new HashSet<Integer>();
   
   /** Saved interactive frame state. */
   private FrameState savedState = null;

   /** Mark frame as such that need focus reset on trigger exit. */
   private boolean focusReset = false;
   
   /** 
    * Mark frame as such that need to honor next conditional down and handle
    * it as unconditional. 
    */
   private boolean honorNextConditionalDown = false;

   /** The scrollable container to place inside the ScrollPane. */
   protected ScrollContainer contentPane = null;
   
   /** Scroll pane which is actual container. */
   protected ScrollPane frameScroll = null;
   
   /** Frame attribute container. */
   private FrameConfig config = null; 

   /** Down frame flag */
   protected boolean down = false; 

   /** Down frame cursor position: 0 - top, maxDown - bottom */
   protected int cursor = -1; 

   /** Down frame headers height; recalculated dynamically  */
   protected int headersHeight = 0; 
   
   /** Down frame vertical step; recalculated dynamically */
   private int verticalStep = 0; 

   /** Vertical size of iteration without trailing SKIPs. */
   protected int verticalStepBoxed = 0;
   
   /** Content of the down frame body */
   protected Widget[][] downBody = null;

   /** Underline flags of the components in the down frame body */
   protected boolean[][] underline = null;

   /** Entire row underline flag. */
   protected boolean[] underlinedRow = null; 
   
   /** Iteration is touched flag. */
   protected boolean[] touched = null; 

   /** Iteration was enabled and need to be cleared flag. */
   private boolean[] needClear = null; 
   
   /** Starting line numbers of every iteration/underline */
   protected int[] startRow = null; 
   
   /** Components which do not belong to down frame body. */
   protected Widget[] nonbody = null;
   
   /** Saved DOWN value. */
   private int saveDown = -1;

   /**
    * Last row in the down body. It may differ from the DOWN value if some
    * components are underlined.
    */
   protected int lastRow = -1;

   /** Honor trailing skips (only if there are regular widgets too). */
   protected boolean trailingSkips = false; 

   /** Streams that have had headers output. */
   protected Set<Stream> headersOutput = null;

   /** Set to true if frame has fixed (explicitly specified) width. */
   private boolean fixedWidth = false;

   /** 
    * Track case when "need clear" is signalled before down body is 
    * initialized. 
    */
   private boolean pendingNeedClear = false;
   
   /** Forced peer frame refresh on next VIEW call. */
   private boolean forcedRefresh = false;

   /** Maps resized editors to their original heights. */
   private Map<Editor, Integer> resizH = new HashMap<Editor, Integer>();

   /** Maps relocated widgets to their original y-coordinates. */
   private Map<Widget, Integer> relocY = new HashMap<Widget, Integer>();

   /** Keeps the original frame height. */
   private int origHeight = 0;

   /** Storage for 'spare' iteration values. */
   private SpareIteration spareIteration = null;

   /** Mark pending cursor wrap state. */
   private boolean pendingCursorWrap = false;

   /** New cursor position. */
   private int pendingCursorPos = 0;

   /** Intermediate value for cursor when cursor wrap is pending. */
   private int newRow = 0;

   /**
    * Special mode in which any VIEW call clears the current iteration and
    * any partial update clears the whole iteration. 
    */
   private boolean clearMode = false;
   
   /**
    * For down and fixed-height frames, this indicated down must be computed
    * later.
    */
   private boolean pendingFixedDown = false;

   /**
    * The HEIGHT-CHARS Progress attribute that was specified upon frame
    * creation.
    */
   private int originalHeightChars = 0;

   /** Size of leading SKIP widgets in iteration. */
   private int leadingSkipSize  = -1;

   /** Size of trailing SKIP widgets in iteration. */
   private int trailingSkipSize = -1;

   /** Helper class which encapsulates all focus transfer logic. */
   private FrameFocusTransferManager focusTransferManager = 
         new FrameFocusTransferManager(this);

   /** CHOOSE processing handler. */
   protected ChooseHandler chooseHandler = new ChooseHandler(this);
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public Frame(ComponentConfig cfg)
   {
      boolean save_visible = cfg.isVisibleDirect();
      cfg.setVisibleDirect(false);
      
      // do not use default constructor
      forceVisibility(false);
      enable();
      setLocation(BaseConfig.INV_COORD, BaseConfig.INV_COORD);
      setInsets(new Insets(0, 0, 0, 0));
      
      originalHeightChars = cfg.getHeightChars();

      if (cfg.getHeightChars() != 0 &&
          cfg.getWidthChars() != 0 &&
          cfg.getDown() == 0)
      {
         pendingFixedDown = true;
      }
      
      config = new FrameConfig(this, cfg);
      contentPane = new ScrollContainer(((FrameConfig) cfg).isScrollable());
      frameScroll = screen().getFactory().createScrollPane(contentPane);
      add(frameScroll);
      
      fixedWidth = (cfg.getWidthChars() > 0);
      config.applyConfig(cfg);

      frameScroll.setLocation(0, 0);

      if (cfg.getDown() != 1)
         down = true;
      
      savedDown  = cfg.getDown();
      
      config.setType(ComponentConfig.FRAME);
      
      if (save_visible)
      {
         ThinClient.placeFrame(this);
         config.setVisible(save_visible);
      }
   }

   /**
    * Check if at least one component able to accept input is enabled.
    * 
    * @return   <code>true</code> if frame has enabled components and
    *           <code>false</code> otherwise.
    */
   public boolean hasEnabled()
   {
      for (Widget comp : getContentPane().widgets())
      {
         if (comp instanceof Label)
            continue;
         if (comp instanceof Skip)
            continue;
         if (comp instanceof Space)
            continue;
         if(comp.isEnabled())
            return true;
      }
      
      return false;
   }

   /**
    * Check if at least one component of any type (including labels) is
    * enabled.
    * 
    * @return   <code>true</code> if frame has enabled components and
    *           <code>false</code> otherwise.
    */
   public boolean hasAnyEnabled()
   {
      for (Widget comp : getContentPane().widgets())
      {
         if(comp.isEnabled())
            return true;
      }
      
      return false;
   }
   
   /**
    * Set hard format check state.
    * 
    * @param    formatCheck
    *           New state of format check.
    */
   public void setFormatCheck(boolean formatCheck)
   {
      for (Widget comp : getContentPane().widgets())
      {
         if (comp instanceof FillIn)
            ((FillIn) comp).setFormatCheck(formatCheck);
      }
   }

   /**
    * This method declares the Frame container valid for clipping of PAINT
    * events originated within it.
    * 
    * @return   <code>true</code> because Frame allows clipping.
    */
   @Override
   public boolean clipping() 
   {
      return true;
   }

   /**
    * Get clipping rectangle.
    *
    * @return   clipping rectangle which takes into account frame border.
    */
   @Override
   public Rectangle getClipRect()
   {
      Rectangle rect  = super.getClipRect();
      FrameConfig cfg = config;

      rect = new Rectangle(rect.top()    - cfg.getBorderTopChars(),
                           rect.left()   - cfg.getBorderLeftChars(),
                           rect.bottom() - cfg.getBorderBottomChars(),
                           rect.right()  - cfg.getBorderRightChars());

      return rect;
   }
   
   /**
    * Requesting the focus to the <code>Frame</code>.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled())
         return;
      
      super.requestFocus();
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public FrameConfig config()
   {
      return config;
   }

   /**
    * Recalculate frame layout.
    */
   public void relayout()
   {
      if (autoLayout)
         doLayout();
   }
   
   /**
    * The methods calls the LayoutManager service for the widgets which are
    * not yet positioned explicitly. The widgets which were added to the frame
    * with the specified position are not involved in this action.
    */
   @Override
   public void doLayout()
   {
      doLayout(false);
   }
   
   /**
    * Process frame layout.
    * 
    * @param   suppressDrawing
    *          If <code>true</code> then drawing is suppressed even if frame
    *          is visible.
    */
   public void doLayout(boolean suppressDrawing)
   {
      if (configUpdateActive)
         return;
      
      startSizeChange(savedWidgets != null);

      Dimension d = super.dimension();
      
      if (savedWidgets == null)
         saveFrameDetails();
      else
         d = restoreFrameDetails();
      
      Dimension dMin = getLayout().minimumSize(createTempContainer());
      int titleLen = config.getTitle().length();
      boolean readjust = false;
      
      if (d.width == 0 || d.height == 0)
      {
         int saveWidth = calculateWidth(d, titleLen);
         
         d = dMin;
         
         readjust = calculateDownAndSetSize(d, titleLen, saveWidth);
      }
      else
      {
         setInnerAndOuterSize(d, dMin, titleLen);
      }

      getLayout().doLayout(createTempContainer());
      
      if (readjust)
         adjustDown();
      else
         fillDownBody();
      
      detectTrailingSkips();

      endSizeChange(isVisible() && !isRedirected() && !suppressDrawing);
   }

   /**
    * Create temporary widget container for layout processing.
    * 
    * @return  instance of {@link TempContainer}.
    */
   private TempContainer createTempContainer()
   {
      return new TempContainer(savedWidgets, getContentPane(), this);
   }

   /**
    * Restore frame details.
    * 
    * @return  frame dimension.
    */
   private Dimension restoreFrameDetails()
   {
      config.setDown(savedDown);
      super.setSize(savedDim);
      getContentPane().setSize(savedPaneDim);
      restoreWidgetData(savedWidgets);
      
      return savedDim.copy();
   }

   /**
    * Save frame details.
    */
   private void saveFrameDetails()
   {
      savedWidgets = saveWidgetData();
      savedDim = new Dimension(config.getWidthChars(), config.getHeightChars());
      super.setSize(savedDim);
      savedPaneDim = getContentPane().dimension();
   }

   /**
    * Calculate frame height delta.
    * 
    * @return height delta.
    */
   private int calculateHeightDelta()
   {
      return config.getBorderTopChars() + config.getBorderBottomChars();
   }

   /**
    * Calculate frame width delta.
    * 
    * @return width delta.
    */
   protected int calculateWidthDelta()
   {
      return config.getBorderLeftChars() + config.getBorderRightChars();
   }

   /**
    * Calculate frame width.
    * 
    * @param   d
    *          Initial frame dimensions.
    * @param   titleLen
    *          Length of frame title.
    *          
    * @return  final frame width.
    */
   private int calculateWidth(Dimension d, int titleLen)
   {
      int deltaW = calculateWidthDelta();
      
      if (d.width > 0 && (titleLen + deltaW) > d.width)
         d.width = titleLen + deltaW; 
      
      int saveWidth  = d.width;
      autoLayout = true;
      
      // trigger recalculation of frame size

      // force 
      if (saveWidth > deltaW)
         setSize(saveWidth - deltaW, d.height);
      return saveWidth;
   }

   /**
    * Calculate frame DOWN value and adjust frame size accordingly.
    * 
    * @param   d
    *          Initial frame dimensions.
    * @param   titleLen
    *          Title length.
    * @param   saveWidth
    *          Saved frame width.
    *          
    * @return  <code>true</code> if frame needs readjustment.
    */
   private boolean calculateDownAndSetSize(Dimension d, 
                                           int titleLen, int saveWidth)
   {
      int deltaW = calculateWidthDelta();
      int deltaH = calculateHeightDelta();
      boolean readjust = false;
      
      // redirected non-boxed frames for which a display statement has only  
      // skip widgets should have width 1
      if (d.width <= 0 && isRedirected() && !config.isBox())
      {
         d.width = 1;
      }

      if (d.width < titleLen)
         d.width = titleLen;
      
      if (config.getDown() == 0)
      {
         config.setDown(1);
         readjust = true;
      }
      
      if (d.height == 0 && config.getDown() > 1)
      {
         // this happens only of frame has no widgets at all
         d.height = config.getDown();
      }
      else
      {
         if (config.getDown() > 1)
            readjust = true;
      }

      if (saveWidth > deltaW && saveWidth > d.width)
         d.width = saveWidth - deltaW;
      
      d.width  += deltaW;
      d.height += deltaH;

      setSize(d.width, d.height);
      return readjust;
   }

   /**
    * Set size of frame components.
    * 
    * @param   d
    *          Initial dimensions.
    * @param   dMin
    *          Minimal dimensions.
    * @param   titleLen
    *          Title length.
    */
   private void setInnerAndOuterSize(Dimension d, Dimension dMin, int titleLen)
   {
      int deltaW = calculateWidthDelta();
      int deltaH = calculateHeightDelta();

      setSize(d.width, d.height);
      
      dMin.width  += deltaW;
      dMin.height += deltaH;
      
      if (dMin.width > d.width || dMin.height > d.height)
      {
         dMin.width  = Math.max(d.width , dMin.width ) - deltaW; 
         dMin.height = Math.max(d.height, dMin.height) - deltaH;
         
         setInnerSize(dMin.width, dMin.height);
      }
      
      if (titleLen > d.width)
      {
         String s = config.getTitle().substring(0, d.width);
         config.setTitle(s);
      }
   }

   /**
    * Get original frame dimension provided by application.
    * 
    * @return original frame dimension.
    */
   public Dimension getSavedDimension()
   {
      return savedDim;
   }
   
   /**
    * Change the original frame dimension, which was initially
    * provided by application.
    * 
    * @param   width
    *          The width of the frame.
    * @param   height
    *          The height of the frame.
    */
   public void setSavedDimension(int width, int height)
   {
      savedDim = new Dimension(width, height);
   }
   
   /**
    * Set the new size of the frame.
    * 
    * @param   width
    *          The new value for the frame width.
    * @param   height
    *          The new value for the frame height.
    */
   @Override
   public void setSize(int width, int height)
   {
      boolean activate = isConfigUpdateActive();
      if (activate)
         startSizeChange(false);
      setSizeImpl(width, height);
      if (activate)
         endSizeChange(false);
   }
   
   /**
    * Set the new width of the frame.
    * 
    * @param   width
    *          The new value for the frame width.
    */
   @Override
   public void setWidth(int width)
   {
      setSizeImpl(width, super.dimension().height);
   }

   /**
    * Set the new height of the frame.
    * 
    * @param   height
    *          The new value for the frame height.
    */
   @Override
   public void setHeight(int height)
   {
      setSizeImpl(super.dimension().width, height);
   }

   /**
    * Adding new widget to the container. This method does not require the
    * coordinates to be defined. The location will be selected automatically.
    * 
    * @param   widget
    *          The new widget to be added to the container.
    */
   public void addWidget(Widget widget)
   {
      contentPane.add(widget);

      if (widget instanceof Editor)
      {
         withEditors = true;
      }
   }

   /**
    * Draw this container.
    */
   @Override
   public void draw()
   {
      // skip drawing if current system mode does not match frame mode 
      if (redirected != ThinClient.getInstance().isRedirected())
         return;
      
      if (isRedirected() != isPlacedForRedirected())
      {
         // frame placement does not match current frame state (redirected or
         // interactive) so we can't draw frame correctly and ought to ignore
         // the request
         return;
      }
      
      // draw the enclosing frame (but only if the insets are nonzero)
      Point origReal = new Point(location());
      Dimension sizeReal = size().copy();
      
      sizeReal.width -= calculateWidthDelta();

      if (sizeReal.height <= 0)
         return;

      // In interactive mode drawing of empty frames is suppressed.
      // Also, since such a frame is not displayed, it should not affect
      // pausing.
      if (!isRedirected() &&  sizeReal.width <= 0)
      {
         needPause = false;
         return;
      }
      
      draw(origReal, sizeReal);
      
      chooseHandler.drawIfActive();
   }
   
   /**
    * Returns the handle of the widget with number inside the frame.
    * 
    * @param   number
    *          The widget number whose handle should be returned.
    * 
    * @return  The selected component with the given number.
    */
   public Widget getSelectedWidget(int number)
   {
      return getContentPane().get(number);
   }

   /**
    * Moves this frame to the bottom of other widgets in the Tab ordering.
    * 
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   public boolean moveToBottom()
   {
      if (parent() == null)
         return false;

      parent().moveToBottom(this);
      return true;
   }

   /**
    * Moves this frame to the top of other widgets in the Tab ordering.
    * 
    * @return  <code>true</code> if operation was successful and
    *          <code>false</code> otherwise.
    */
   public boolean moveToTop()
   {
      if (parent() == null)
         return false;

      parent().moveToTop(this);
      return true;
   }
   
   /**
    * Mark this frame as streamed.
    * This marking works in way only.
    */
   public void markAsStreamed()
   {
      streamed = true;
   }

   /**
    * Returns <code>true</code> if frame is in active use on the server and
    * <code>false</code> if the frame is not in use by the server and can
    * be destroyed.
    * 
    * @return   <code>true</code> if frame is in scope.
    */
   public boolean isAlive()
   {
      return live;
   }
   
   /**
    * Mark start and end of frame lifetime at client.
    * 
    * @param    alive
    *           New value for the flag.
    */
   public void setAlive(boolean alive)
   {
      this.live = alive;
   }

   /**
    * Returns <code>true</code> if frame is in frame scope.
    * 
    * @return   <code>true</code> if frame is in scope.
    */
   public boolean isInScope()
   {
      return inScope;
   }

   /**
    * Mark start and end of frame scope.
    * 
    * @param    inScope
    *           New value for the flag.
    */
   public void setInScope(boolean inScope)
   {
      this.inScope = inScope;
      
      // when frame leaves scope, no-hide option should be cleared
      if (!inScope)
      {
         config.setNoHide(false);
         viewed = false;
      }

      // when frame enters scope, restore the no-hide status
      if (inScope)
      {
         config.setNoHide(config.isNoHideConfig());
         streamed = false;
         setFormatCheck(false);
      }
   }
   
   /**
    * Handle change of frame visibility.
    * 
    * @param   visible
    *          New visibility state. <code>true</code>- show frame,
    *          <code>false</code>- hide frame.
    */
   @Override
   public void setVisible(boolean visible)
   {
      forcePlaceFrame(visible);
      
      changeVisibility(visible);
      
      if (!visible)
      {
         onFromVisibleToHidden();
      }
      else
      {
         onFromHiddenToVisible();
      }
   }

   /**
    * Process transition from hidden to visible state.
    */
   private void onFromHiddenToVisible()
   {
      active = true;
      viewed = true;
   }

   /**
    * Process transition from visible to hidden state.
    */
   private void onFromVisibleToHidden()
   {
      resetPlacementQuirk(false);
      
      handleFlags();

      handleDynamicDown();
      
      handleDynamicLocation();

      // frame is hidden, remove it from all relevant places 
      if (!isAlive() && !protect)
         ThinClient.removeFrame(this);
      
      resetForcedRefresh();
      
      setFlushed(false);
      
      if (!inScope)
         viewed = false;
      
      forceWipeScreen = false;
   }

   /**
    * Change widget visibility.
    * 
    * @param   visible
    *          New visibility state.
    */
   private void changeVisibility(boolean visible)
   {
      boolean handleBody = (visible != isVisible());
      
      // frame is about to appear or disappear; clean up
      // any put screen output within the rectangle occupied by frame
      if (handleBody) 
         refreshOnVisibilityChange(visible);

      super.setVisible(visible);
      
      if (handleBody)
         setBodyVisible(visible);
   }

   /**
    * Force placement of frame at the screen.
    * 
    * @param   visible
    *          New visibility state.
    */
   private void forcePlaceFrame(boolean visible)
   {
      if (visible &&
          (location().x == BaseConfig.INV_COORD || 
           location().y == BaseConfig.INV_COORD))
      {
         ThinClient.placeFrame(this);
      }
   }

   /**
    * Handle dynamic DOWN changes.
    */
   private void handleDynamicDown()
   {
      if (dynamicDown && pendingDown != -1)
      {
         savedDown = pendingDown;
         config.setDown(pendingDown);
         pendingDown = -1;
      }
   }

   /**
    * Handle various frame flags.
    */
   private void handleFlags()
   {
      focusReset = false;
      
      if (canHide)
         focusId = -1;

      canHide = false;
   }

   /**
    * Handle dynamic frame location changes.
    */
   private void handleDynamicLocation()
   {
      Point loc = location();

      if (dynamicCol != BaseConfig.INV_COORD && dynamicCol != loc.x)
      {
         loc.x = dynamicCol;
      }
      if (dynamicRow != BaseConfig.INV_COORD && dynamicRow != loc.y)
      {
         loc.y = dynamicRow;
      }
      
      setLocation(loc);
   }

   /**
    * Handle refreshing of parts of frame on visibility changes.
    * 
    * @param    visible
    *           New visibility state.
    */
   private void refreshOnVisibilityChange(boolean visible)
   {
      Dimension size = dimension();
      boolean box = config.isBox();
      
      if (box && (size.width < 3 || size.height < 3))
         box = false;
      
      // refresh the rectangle with the double buffered driver
      // rectangle includes the box for view but not for hide
      
      ThinClient client = ThinClient.getInstance();
      
      if (!visible)
      {
         screen().refresh(calculateFrameRect(visible));
      }
      else 
      { 
         boolean hasTitle = hasTitle();
         
         Rectangle r = calculateFrameRect(true);

         if (hasTitle && size.width > 3)
         {
            refreshFrameArea(client, r);
         }

         if (!viewed)
         {
            boolean refresh = checkRefreshNecessary(r);
            
            if (refresh && box)
            {
               refreshBoxedFrame(client, r);
            }
         }
      }
      
      if (!visible && isTopFrame() && !isRedirected() && box)
      {
         // hiding the top frame clears PUT SCREENs to the right
         // of the frame on all unoccupied lines
         
         refreshOnTopFrameHide(client);
      }
   }

   /**
    * Process screen refreshing when frame is top frame in Z-order.
    *  
    * @param   client
    *          {@link ThinClient} reference.
    */
   private void refreshOnTopFrameHide(ThinClient client)
   {
      Rectangle r = calculateFrameRect(true);
      
      r = new Rectangle(r.top(), 
                        r.right(),
                        r.bottom(), 
                        ThinClient.getScreenSize().width - 1);
      
      Iterable<Rectangle> list = getUnoccupied(r);
      
      for (Rectangle rect : list)
      {
         client.conditionalRefresh(rect);
      }
   }

   /**
    * 
    * @param   client
    *          {@link ThinClient} reference.
    * @param   rect
    *          Initial frame rectangle.
    */
   private void refreshBoxedFrame(ThinClient client, Rectangle rect)
   {
      boolean overlay = config.isOverlay();
      
      // refresh sides
      Rectangle top = new Rectangle(rect.top(), rect.left(), 
                                    rect.top(), rect.right());
      
      Rectangle bottom = new Rectangle(rect.bottom(), rect.left(), 
                                       rect.bottom(), rect.right());

      Rectangle left = new Rectangle(rect.top() + 1   , rect.left(), 
                                     rect.bottom() - 1, rect.left());
      
      Rectangle right = new Rectangle(rect.top() + 1   , rect.right(), 
                                      rect.bottom() - 1, rect.right());
      
      if (overlay)
      {
         client.conditionalLineRefresh(top);
         
         if (ThinClient.isRefreshQuirkActive())
            screen().refresh(bottom);
         else
            client.conditionalLineRefresh(bottom);
      }
      else
      {
         screen().refresh(top);
         screen().refresh(bottom);
      }
      
      screen().refresh(left);
      screen().refresh(right);
   }

   /**
    * Check if refreshing is actually necessary.
    * 
    * @param   rect
    *          Frame rectangle.
    *          
    * @return  <code>true</code> if refreshing is necessary.
    */
   private boolean checkRefreshNecessary(Rectangle rect)
   {
      if (intersectsTopFrame())
      {
         screen().refresh(calculateFrameRect(false));
         return true;
      }
      
      return UiUtils.containsFrames(this, rect);
   }

   /**
    * Refresh area which belongs to frame.
    * 
    * @param   client
    *          Reference to {@link ThinClient}.
    * @param   rect
    *          Frame rectangle.
    */
   private void refreshFrameArea(ThinClient client, Rectangle rect)
   {
      boolean overlay = config.isOverlay();

      Rectangle refreshRect = new Rectangle(rect.top(), rect.left()  + 1, 
                                            rect.top(), rect.right() - 1);
      if (overlay)
      {
         client.conditionalLineRefresh(refreshRect);
      }
      else
      {
         screen().refresh(refreshRect); 
      }
   }

   /**
    * Check if frame has title.
    * 
    * @return   <code>true</code> if frame has valid non-empty title.
    */
   protected boolean hasTitle()
   {
      if (config.getTitle() != null &&
          config.getTitle().length() > 0)
      {
         return true;
      }
      
      return false;
   }

   /**
    * Reset automatic frame placing flag. 
    */
   public void resetAutoplacing()
   {
      autoplacing = false;
   }

   /**
    * Get current state of the automatic frame placing flag. If automatic
    * placing is active frame coordinates are automatically assigned by frame
    * manager when frame is displayed.
    * 
    * @return  <code>true</code> if automatic placing is active for the
    *          frame.
    */
   public boolean isAutoplacing()
   {
      return autoplacing;
   }
   
   /**
    * Check status of active frame scope.
    * 
    * @return  Current status of the flag.
    */
   public boolean isActive()
   {
      return active;
   }
   
   /**
    * Check the streamed flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isStreamed()
   {
      return streamed;
   }
   
   /**
    * Check if frame is a <code>Dialog</code>.
    * 
    * @return  <code>true</code> if frame is a <code>Dialog</code>.
    */
   public boolean isDialog()
   {
      return config.isDialog();
   }

   /**
    * Check if frame is a top frame in the parent's Z-order.
    * 
    * @return  <code>true</code> if frame is top frame.
    */
   //TODO: duplicates other similar methods?
   public boolean isTopFrame()
   {
      Widget[] c = parent().widgets();
      
      for (int i = (c.length - 1); i >= 0; i--)
      {
         if (c[i] instanceof Frame)
            return c[i] == this;
      }
      
      return false;
   }

   /**
    * Get frame DOWN attribute value.
    * 
    * @return  current attribute value.
    */
   public int getDown()
   {
      if (!isVisible())
         return 0;
      
      return config.getDown();
   }
   
   /**
    * Get the text value of the currently focused widget taking into account
    * the last CHOOSE operation.
    *
    * @param    widget
    *           widget to query the value for or <code>null</code> which
    *           indicates the currently focused widget, if any
    * @return   The data from the currently focused widget or "" if there is
    *           no data or on any error.
    */
   public FrameData getFrameValue(Widget widget)
   {
      return isDown() ? getDownFrameValue(widget) : getRegularFrameValue(widget);
   }

   /**
    * Get frame value for regular (1-DOWN) frames.
    * 
    * @param   widget
    *          
    * @return  obtained value.
    */
   private FrameData getRegularFrameValue(Widget widget)
   {
      int id = -1;
      String result = "";
      
      Widget comp = locateRegularBodyComponent(widget);
      
      if (comp instanceof DataContainer) 
      {
         if (comp instanceof FillIn) 
            result = ((FillIn) comp).getText(true);
         else
            result = ((DataContainer) comp).getText();

         id = UiUtils.lookupWidgetId(comp);
      }
      
      return new FrameData(result, id);
   }

   /**
    * Get frame value for DOWN frames.
    * 
    * @param   widget
    *          
    * @return  obtained value.
    */
   private FrameData getDownFrameValue(Widget widget)
   {
      int id = -1;
      String result = "";

      Widget comp = locateDownBodyComponent(widget);
      
      int ndx = (chooseHandler.index() >= 0) ? chooseHandler.index() : 0;

      if (comp instanceof DataContainer)
      {
         result = getFrameValue(ndx, comp);
         id = UiUtils.getWidgetId(comp);
      }
      return new FrameData(result, id);
   }

   /**
    * Locate actual component for regular (1-DOWN) frames.
    * 
    * @param   comp
    *          Component to locate.
    *           
    * @return  located component.
    */
   private Widget locateRegularBodyComponent(Widget comp)
   {
      if (comp == null)
         comp = super.currentFocus();

      if (comp instanceof SelectionListBody)
      {
         // find the selection list
         while (!(comp instanceof SelectionList) && comp != null)
            comp = comp.parent();
      }
      return comp;
   }

   /**
    * Locate actual component for DOWN frames.
    * 
    * @param   comp
    *          Component to locate.
    *           
    * @return  located component.
    */
   private Widget locateDownBodyComponent(Widget comp)
   {
      if (comp == null)
         comp = chooseHandler.focusedCol();
      
      if (comp == null)
         comp = super.currentFocus();

      if (comp instanceof SelectionListBody)
      {
         while (!(comp instanceof SelectionList))
            comp = comp.parent();                      
      }
      return comp;
   }

   /**
    * Return the selected component (using the CHOOSE index) in a down frame
    * that corresponds with the given base component.
    *
    * @param    base
    *           The base component of the frame.
    *
    * @return   The corresponding component in the same column and at the
    *           selected row if this is a down frame (and in ROW mode).
    *           The passed component will be returned if the frame is not a
    *           down frame or if the choose is not in ROW mode).
    */
   public Widget getSelectedField(Widget base)
   {
      boolean indirect = (!chooseHandler.isFieldMode() && isDown());
      return indirect ? getField(chooseHandler.index(), base) : base;
   }

   /**
    * Get current state of canHideOthers flag.
    * 
    * @return  value of canHideOthers flag.
    */
   public boolean isCanHideOthers()
   {
      return canHideOthers;
   }

   /**
    * Set current state of canHideOthers flag.
    * 
    * @param   canHideOthers
    *          New value of the flag.
    */
   public void setCanHideOthers(boolean canHideOthers)
   {
      this.canHideOthers = canHideOthers;
   }

   /**
    * Get state of canWipeScreen flag.
    * 
    * @return  value of canWipeScreen flag.
    */
   public boolean isCanWipeScreen()
   {
      return canWipeScreen;
   }

   /**
    * Set state of canWipeScreen flag.
    * 
    * @param   canWipeScreen
    *          New value of the flag.
    */
   public void setCanWipeScreen(boolean canWipeScreen)
   {
      this.canWipeScreen = canWipeScreen;
   }

   /**
    * Set state of forceWipeScreen flag.
    * 
    * @param   forceWipeScreen
    *          New value of the flag.
    */
   public void setForceWipeScreen(boolean forceWipeScreen)
   {
      this.forceWipeScreen = forceWipeScreen;
   }
   
   /**
    * Return current state of 'force wipe screen' flag.
    * 
    * @return   <code>true</code> if frame must wipe screen on next VIEW call.
    */
   public boolean isForceWipeScreen()
   {
      return forceWipeScreen;
   }
   
   /**
    * Get state of canHide flag.
    * 
    * @return  value of canHide flag.
    */
   public boolean isCanHide()
   {
      return canHide;
   }

   /**
    * Set current state of canHide flag.
    * 
    * @param   canHide
    *          New value of the flag.
    */
   public void setCanHide(boolean canHide)
   {
      this.canHide = canHide;
      this.active = !canHide;
   }

   /**
    * Adjust down frame height according to current conditions.
    * <p>
    * Note. In redirected mode, down frame possibly needs just one iteration
    * of data ever. The current implementation doesn't detect this case and
    * adjusts the frame based on the screen dimensions, which is incorrect.
    * The UI would benefit from adjusting the down frame to one iteration as
    * it minimizes the consumed resources.
    * TO-DO.
    */
   public void adjustDown()
   {
      if (!isDownCalculated())
      {
         fillDownBody();
      }

      int delta = 0;
      
      if (isDynamic())
      {
         Point location = screenLocation();
         if (location.y < 0 || location.x < 0)
         {
            // frame was not yet placed, do not adjust down
            return;
         }

         // calculate size automatically
         delta = calculateDown(location);
      }

      Dimension d = super.dimension();

      calculateHeight(d, delta);

      getContentPane().setHeight(d.height);

      d.height += calculateHeightDelta();

      setSize(d.width, d.height);

      fillDownBody();
   }

   /**
    * Calculate frame height.
    * @param   d
    *          Frame dimensions.
    * @param   delta
    *          Height delta.
    */
   private void calculateHeight(Dimension d, int delta)
   {
      if (getOriginalHeightChars() != 0)
      {
         // If the frame height was explicitly specified in the Progress code,
         // we should use it.
         d.height = getOriginalHeightChars() - calculateHeightDelta();
      }
      else
      {
         // Calculate the frame height using height of fixed part and height 
         // of iterations.
         d.height = getFixedHeight() +
                    (getIterationHeight() * config.getDown()) + delta;
      }
   }

   /**
    * Calculate DOWN value.
    * 
    * @param   location
    *          Frame location.
    *          
    * @return  extra frame height row count.
    */
   private int calculateDown(Point location)
   {
      int delta = 0;
      int sz = 0;

      if (isPendingFixedDown())
      {
         // compute the fixed down from the fixed height
         sz = config.getHeightChars();
      }
      else
      {
         sz = ThinClient.getInstance().getScreenLines();
         sz -= location.y + getFixedHeight();
      }
      
      sz -= calculateHeightDelta();

      if (isFixedRow)
         delta = sz % getIterationHeight();

      sz /= getIterationHeight();

      if (sz < 1)
         sz = 1;

      config.setDown(sz);
      
      if (isPendingFixedDown())
      {
         this.savedDown = sz;
         setPendingFixedDown(false);
         if (sz == 1)
         {
            // mark it as non-down, even if it is fixed
            setDown(false);
         }
      }
      return delta;
   }

   /**
    * Returns <code>true</code> if size of the DOWN frame should be
    * dynamically recalculated.
    * 
    * @return  <code>true</code> if size should be recalculated.
    */
   public boolean isDynamic()
   {
      return (savedDown == 0);
   }

   /**
    * Get original DOWN attribute value.
    * 
    * @return  The value of the attribute.
    */
   public int getSaveDown()
   {
      return savedDown;
   }

   /**
    * Return minimal frame height. For DOWN(0) frames this is height of the
    * fixed part of the frame and height of one iteration.
    * 
    * @return  Minimal frame height.
    */
   public int getMinHeight()
   {
      if (!isDynamic())
         return height();
      
      return getIterationHeight() + getFixedHeight() + calculateHeightDelta();
   }

   /**
    * Returns state of the <code>needPause</code> flag.
    * 
    * @return  Returns the current value of the flag.
    */
   public boolean isNeedPause()
   {
      return needPause;
   }

   /**
    * Directly set {@link #needPause} flag.
    * 
    * @param   needPause
    *          New flag value.
    */
   public void setNeedPauseDirect(boolean needPause) 
   {
      this.needPause = needPause;
   }
   
   /**
    * Sets <code>needPause</code> flag to specified state.
    * 
    * @param   needPause
    *          New value for the flag.
    */
   public void setNeedPause(boolean needPause)
   {
      // force NO-PAUSE for redirected frames.
      if (ThinClient.getInstance().isRedirected())
         needPause = false;
      
      this.needPause = needPause;
      
      if (!needPause)
         setForcePause(needPause);
   }
   
   /**
    * Frame-specific version of the <code>toString()</code>.
    * 
    * @see  java.lang.Object#toString()
    */
   @Override
   public String toString()
   {
      return shortName() + 
            " " + config.getID() + 
            " " + size() +
            " " + location() +  
            " visible " + isVisible() +
            " can " + canHide + 
            "; ";
   }
   
   /**
    * Debugging Aid.
    * Visualizes the redirected terminal's contents.
    *
    * @return array of strings that represent the contents of the terminal
    */
   public String[] view()   
   {
      if (!redirected)
         return new String[] {"interactive"};

      Stream st = ThinClient.getInstance().streamFromId(streamId);
      RedirectedTerminal rt = ThinClient.redirectedTerminalForStream(st);
      return rt.view();
   }

   /**
    * Override key processing.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (!chooseHandler.isChooseActive())
      {
         if (event.source() != this)
            super.processKeyEvent(event);
         else
            processKeyListeners(event);

         // stop processing if a listener consumed the event
         if (event.isConsumed())
         {
            return;
         }
         
         int eventID = event.id();

         // process the NEXT-FRAME or PREV-FRAME event
         if (eventID == Keyboard.SE_NEXT_FRAME || 
             eventID == Keyboard.SE_PREV_FRAME)
         {
            boolean next = eventID == Keyboard.SE_NEXT_FRAME;
            ThinClient.getInstance().handleNextPrevFrameEvents(this, next);
         }
         return;
      } 
      chooseHandler.processKeyEvent(event);
   }
   
   /**
    * Dynamic DOWN adjustment.
    * 
    * @param   down
    *          New value of the DOWN attribute.
    */
   public void adjustDown(int down)
   {
      if (down == -1)
         return;

      dynamicDown  = true;
      
      if (down < 0)
         down = 0;
      
      if (isVisible())
      {
         pendingDown = down;
      }
      else
      {
         savedDown = down;
         config.setDown(down);
         relayout();
      }
   }

   /**
    * Set dynamic frame row.
    * 
    * @param   row
    *          Dynamically calculated row.
    */
   public void setDynRow(int row)
   {
      dynamicRow = row;
      
      if (!isVisible() && dynamicRow != BaseConfig.INV_COORD)
      {
         Point loc = location();
         loc.y = row;
         setLocation(loc);
         resetAutoplacing();
      }
   }

   /**
    * Set dynamic column.
    * 
    * @param   column
    *          Dynamically calculated column.
    */
   public void setDynCol(int column)
   {
      dynamicCol = column;
      
      if (!isVisible() && dynamicCol != BaseConfig.INV_COORD)
      {
         Point loc = location();
         loc.x = column;
         setLocation(loc);
         resetAutoplacing();
      }
   }
   
   /**
    * Return <code>true</code> if last CHOOSe operation was finished by
    * internal GO event.
    * 
    * @return  Returns current state of <code>internalGo</code> flag.
    */
   public boolean isInternalGo()
   {
      return chooseHandler.isInternalGo();
   }
   
   /**
    * Mark frame as such that requires resetting frame widgets when frame will
    * be displayed.
    * 
    * @param   needReset
    *          <code>true</code> - clear frame on next displaying.
    */
   public void setNeedReset(boolean needReset)
   {
      resetPending = needReset;
   }

   /**
    * Return state of the <code>resetPending</code> flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isNeedReset()
   {
      return resetPending;
   }
   
   /**
    * Mark frames that requires a repaint when will be displayed.
    * 
    * @param   delayedRepaint
    *          <code>true</code> - repaint the frame on the next display.
    */
   public void setDelayedRepaint(boolean delayedRepaint)
   {
      this.delayedRepaint = delayedRepaint;
   }

   /**
    * Return state of the <code>delayedRepaint</code> flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isDelayedRepaint()
   {
      return delayedRepaint;
   }

   /**
    * Get the scope nesting level.
    * 
    * @return  the scope nesting level
    */
   public int getScopeNesting()
   {
      return scopeNesting;
   }

   /**
    * Set the scope nesting level.
    * 
    * @param   scopeNesting
    *          The new scope nesting level
    */   
   public void setScopeNesting(int scopeNesting)
   {
      this.scopeNesting = scopeNesting;
   }

   /**
    * Sets or resets the ALL CLEAR flag for the frame.
    * This flag tells the VIEW/DISPLAY to ignore conditional UPs and DOWNs.
    * 
    * @param   allClear
    *          <code>true</code> - new value for the ALL CLEAR flag
    */
   public void setAllClear(boolean allClear)
   {
      this.allClear = allClear;
   }

   /**
    * Returns the state of the ALL CLEAR flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isAllClear()
   {
      return allClear;
   }
   
   /**
    * Return state of NO-HIDE flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isNoHide()
   {
      return config.isNoHide();
   }
   
   /**
    * Checks whether frame has editors.
    * 
    * @return  <code>true</code> if frame has at least one editor
    */
   public boolean hasEditors()
   {
      return withEditors;
   }
   
   /**
    * Reset frame to its initial state.
    *
    */
   public void reset()
   {
      if (!resetPending)
         return;

      if (isDown())
      {
         resetCursor();
      }

      clearWorker(true);
      resetPending = false;
   }
   
   /**
    * Controllable object repainting.
    */
   @Override
   public void repaint() 
   {
      if (paintable && !configUpdateActive)
         super.repaint();
   }

   /**
    * Set ID of component which will receive focus next time.
    * 
    * @param   focusId
    *          Component ID.
    */
   public void setFocusId(int focusId)
   {
      this.focusId  = focusId; 
   }

   /**
    * ID of component which will receive focus next time.
    * 
    * @return  The component ID.
    */
   public int getFocusId()
   {
      return focusId; 
   }
   
   /**
    * Gets redirection flag.
    * 
    * @return  <code>true</code> if frame is redirected.
    */
   public boolean isRedirected()
   {
      return redirected;
   }
   
   /**
    * Gets the last used stream ID.
    * 
    * @return  the stream ID of the last used stream.
    */
   public int getStreamId()
   {
      return streamId;
   }

   /**
    * Restores tab order that corresponds initial components layout. 
    */
   public void restoreInitialTabOrder()
   {
      if (savedWidgets == null)
         return;

      for (WidgetItem widgetItem : savedWidgets)
      {
         Widget comp = widgetItem.getWidget();

         if (comp.focusTraversable())
            getContentPane().moveToTop(comp);
      }
   }
   
   /**
    * This method is invoked when frame definition is about to be updated.
    * This method is not called when frame is just created first time.
    */
   public void preprocess()
   {
      configUpdateActive = true;
   }

   /**
    * This method is invoked when frame is constructed and configuration of
    * all widgets is set, but no other actions are performed.
    * 
    * @param   suppressRedraw
    *          <code>true</code> means suppress redrawing of frame.
    */
   public void postprocess(boolean suppressRedraw)
   {
      isFixedRow = (config.getRow() != BaseConfig.INV_COORD);
      configUpdateActive = false;
      
      // setPause is set to true only if this is consequence of color change   
      // request, which does not require full redraw, since each widget
      // is updated on screen independently
      if (suppressRedraw && colorChanged)
         setNeedPause(suppressRedraw);
      
      boolean suppressDraw = suppressRedraw || !isVisible();
      
      colorChanged = false;
      
      if (config.isSizeChanged())
      {
         setSavedDimension(config.getWidthChars(), 
                           config.getHeightChars());
         config.setSizeChanged(false);
      }
      
      // finally layout the frame, place widgets, recalculate size
      doLayout(suppressDraw);
      
      if (!suppressRedraw && postponedPlace)
         ThinClient.placeFrame(this);
      if (!suppressRedraw && postponedVisible)
         setVisible(true);
      
      postponedPlace = false;
      postponedVisible = false;
   }
   
   /**
    * Get current state of config update.
    * 
    * @return  <code>true</code> if config is being applied to the widget 
    */
   public boolean isConfigUpdateActive()
   {
      return configUpdateActive;
   }

   /**
    * Enable/disable changing frame visibility after
    * {@link #applyConfig(Widget, ComponentConfig)}.
    * 
    * @param   visible
    *          New state of the flag.
    */
   public void postponeSetVisible(boolean visible)
   {
      postponedVisible = visible;
   }

   /**
    * Enable/disable frame placing after
    * {@link #applyConfig(Widget, ComponentConfig)}.
    * 
    * @param   place
    *          New state of the flag.
    */
   public void postponePlace(boolean place)
   {
      postponedPlace = place;
   }

   /**
    * Set forced pause flag state.
    * 
    * @param   forcePause
    *          New value for the flag.
    */
   public void setForcePause(boolean forcePause)
   {
      this.forcePause = forcePause;
   }
   
   /**
    * Get current state of forced pause flag.
    * 
    * @return  flag state.
    */
   public boolean isForcePause()
   {
      boolean rc = forcePause;
    
      forcePause  = false;
      
      return rc;
   }
   
   /**
    * Notify frame that it is placed.
    */
   public void placed()
   {
      placedForRedirected = redirected;
   }
   
   /**
    * Return true if frame is placement was done for redirected mode.
    * 
    * @return <code>true</code> for redirected mode.
    */
   public boolean isPlacedForRedirected()
   {
      return placedForRedirected;
   }
   
   /**
    * Get current status of 'focus attractor' flag.
    * 
    * @return  current value of the flag.
    */
   public boolean isFocusAttractor()
   {
      return focusAttractor ;
   }
   
   /**
    * Set new value for the 'focus attractor' flag.
    * 
    * @param   attractor
    *          New value for the flag.
    */
   public void setFocusAttractor(boolean attractor)
   {
      focusAttractor = attractor;
   }
   
   /**
    * Get current status of 'visibility changed' flag.
    * 
    * @return  current value of the flag.
    */
   public boolean isVisibilityChanged()
   {
      return visibilityChanged;
   }
   
   /**
    * Set new value for the 'visibility changed' flag.
    * 
    * @param   changed
    *          New value for the flag.
    */
   public void setVisibilityChanged(boolean changed)
   {
      visibilityChanged = changed;
   }
   
   /**
    * Get current status of 'explicitHide' flag.
    * 
    * @return  current value of the flag.
    */
   public boolean isExplicitHide()
   {
      return explicitHide;
   }
   
   /**
    * Set new value for the 'explicitHide' flag.
    * 
    * @param   explicit
    *          New value for the flag.
    */
   public void setExplicitHide(boolean explicit)
   {
      explicitHide = explicit;
   }
   
   /**
    * Notify frame that color of one of its components is changed. 
    */
   void colorChanged()
   {
      if (configUpdateActive)
         colorChanged = true;
   }

   /**
    * Mark/unmark frame as such that need focus reset on trigger exit.
    * 
    * @param    focusReset
    *           <code>true</code> mark frame, <code>false</code> - unmark.
    */
   public void setFocusReset(boolean focusReset)
   {
      this.focusReset = focusReset;
   }
   
   /**
    * Check if frame needs focus reset.
    * 
    * @return   Focus reset flag state.
    */
   public boolean isFocusReset()
   {
      return focusReset;
   }
   
   /**
    * Compute and return the minimum size need to display the widgets within
    * this frame.
    * 
    * @return  The computed minimum size for this frame.
    */
   Dimension getMinimumSize()
   {
      TempContainer tmp = createTempContainer();
      return getLayout().minimumSize(tmp);
   }
   
   /**
    * Get short name of the frame.
    * 
    * @return  Frame short name.
    */
   String shortName()
   {
      String str = config.getName();
      
      if (str == null)
         return "(noname)";
      
      str = str.substring(0, str.indexOf('#'));
      
      return str;
   }

   /**
    * Place frame at top in Z-order, taking into account dialogs. 
    */
   public void placeAtTop()
   {
      // dialog is placed on top unconditionally
      // regular frame is placed under the first visible dialog, if any
      // or else on top
      
      Container parent = parent();
      Widget lst[] = parent.widgets();
      int nlst = lst.length;
      int i = 0;
      boolean dialogs = false;
   
      if (!isDialog())
      {
         for (i = nlst - 1; i >= 0; i--)
         {
            if (lst[i].isVisible() && lst[i] instanceof Frame)
            {
               if (((Frame)lst[i]).isDialog())
               {
                  dialogs = true;
               }
   
               break;
            }
         }
      }
      if (!dialogs)
      {
         moveToTop();
      }
      else
      {
         parent.moveToZpos(this, i);
      }
   }
   
   /**
    * Set the new size of the frame but avoid recursion.
    * 
    * @param   width
    *          The new value for the frame width.
    * @param   height
    *          The new value for the frame height.
    */
   protected void setSizeInt(int width, int height)
   {
      super.setSize(width, height);
   }

   /**
    * Controls object repainting.
    *
    * @param    paintable
    *           <code>true</code> if repaint() calls are to be honored.
    */
   public void setPaintable(boolean paintable)
   {
      this.paintable = paintable;
   }

   /**
    * Protect frame from removal on hide.
    * 
    * @param   protect
    *          New state of the flag;
    */
   public void setProtected(boolean protect)
   {
      this.protect = protect;
   }
   
   /**
    * Make sure that clipping rectangle of redirected frame with trailing
    * SKIPs is set correctly.
    */
   public void touchBody()
   {
      if (isDown() || !isVisible() || !isRedirected())
         return;
      
      repaint();
   }
   
   /**
    * Set redirection flag and the associated stream ID.
    * 
    * @param redirected
    *        state of redirection
    * @param streamId
    *        associated stream ID, if redirected is true.
    */
   public void setRedirected(boolean redirected, int streamId)
   {
      if (redirected && !this.redirected && isVisible())
      {
         savedState  = new FrameState(this);
      }
      
      this.redirected = redirected;
      this.streamId   = streamId;

      // when setting the redirected state for the frame, check if the frame
      // has headers with the stream or not.
      // if it does not, hide it so that the next view() draws headers.
      if (redirected) 
      {
         Stream s = ThinClient.getInstance().streamFromId(streamId);
         if (!isWithHeaders(s))
         {
            forceVisibility(false);
         }
         
         streams.add(Integer.valueOf(streamId));
      }
   }

   /**
    * Check if frame has saved state.
    * 
    * @return   <code>true</code> if frame has saved state.
    */
   public boolean hasSavedState()
   {
      return (savedState != null) && (streams.size() == 0);
   }
   
   /**
    * Restore frame state.
    */
   public void restore()
   {
      if (savedState != null)
         savedState.restore();
   }
   
   /**
    * Unbind stream with given ID.
    *
    * @param    id
    */
   public void unbind(int id)
   {
      if (streams.size() == 0)
         return;

      if (streams.remove(Integer.valueOf(id)))
      {
         if (streams.size() != 0)
            return;

         restore();
      }
   }
   
   /**
    * Get frame flushed state.
    * 
    * @return  Current state of the flag.
    */
   boolean isFlushed()
   {
      return flushed;
   }
   
   /**
    * Set state of <code>flushed</code> flag.
    * 
    * @param   flushed
    *          New state of the flag.
    */
   void setFlushed(boolean flushed)
   {
      this.flushed  = flushed;
   }

   /**
    * Reset placement quirk if necessary.
    * 
    * @param   force
    *          Force resetting of the placement quirk. 
    */
   public void resetPlacementQuirk(boolean force)
   {
      int lastRow = screen().screenDimension().height - Window.RESERVED_LINES;
      
      if (location().y == lastRow && (force || isCanHide() || !isAlive()))
      {
         ThinClient.resetPlacementQuirk();
      }
   }
   
   /**
    * Save initial widget data into array of <code>WidgetItem</code>
    * instances.
    * 
    * @return  Array of saved widget data containers.
    */
   private WidgetItem[] saveWidgetData()
   {
      Widget[] c = getContentPane().widgets();
      ArrayList<WidgetItem> list = new ArrayList<WidgetItem>();
      
      for (int i = 0; i < c.length; i++)
      {
         if (UiUtils.hasConfig(c[i]))
            list.add(new WidgetItem(c[i]));
      }
      
      return list.toArray(new WidgetItem[list.size()]);
   }
   
   /**
    * Restore widget internal data from provided containers.
    * 
    * @param   list
    *          Array of widget containers to restore.
    */
   private void restoreWidgetData(WidgetItem[] list)
   {
      if (list == null)
         return;
      
      for (int i = 0; i < list.length; i++)
      {
         list[i].restore();
      }
   }
   
   /**
    * Calculates the frame's rectangle in screen coordinates
    *
    * @param    withBox
    *           when <code>true</code>, instructs the method to include
    *           the room the frame's box takes, into the rectangle
    * @return   frame's rectangle
    */
   public Rectangle calculateFrameRect(boolean withBox)
   {
      Point loc = screenLocation();
      Dimension size = dimension();

      if (config.isBox() && !withBox)
      {
         loc.x += config.getBorderLeftChars();
         loc.y += config.getBorderTopChars();
   
         size.width -= config.getBorderLeftChars() +
                       config.getBorderRightChars();
         size.height-= config.getBorderTopChars() +
                       config.getBorderBottomChars();
      }

      return new Rectangle(loc, size);
   }

   /**
    * Check if frame intersects top frame.
    * 
    * @return   <code>true</code> if frame intersects top frame in Z-order.
    */
   private boolean intersectsTopFrame()
   {
      Widget[] c = Window.getDefaultWindow().getContentPane().widgets();
      
      for (int i = c.length - 1; i >= 0; i--)
      {
         if (c[i] instanceof Frame)
         {
            Frame frm = (Frame) c[i];
            
            if (frm == this      || 
                !frm.isVisible() || 
                frm.isRedirected() != isRedirected())
            {
               continue;
            }
            
            Rectangle rect = frm.calculateFrameRect(true);

            return rect.intersects(calculateFrameRect(true));
         }
      }
      
      return false;
   }
   
   /**
    * Calculates contiguous rectangles of the given rectangle that have lines
    * not covered with any frame.
    * 
    * @param    rect
    *           original rectangle to inspect
    * 
    * @return   collection of unoccupied rectangles. 
    */
   private Iterable<Rectangle> getUnoccupied(Rectangle rect)
   {
      ArrayList<Frame> frames = new ArrayList<Frame>();
      ArrayList<Rectangle> rects = new ArrayList<Rectangle>();
      
      Window.findFrames(frames, rect, true, isRedirected());
   
      frames.remove(this);
      
      if (frames.isEmpty())
      {
         rects.add(rect);
      }
      else
      {
         for (int i = rect.top(); i <= rect.bottom(); i++)
         {
            frames.clear();
            
            Rectangle test = new Rectangle(i, rect.left(), 
                                           i, rect.right());
            
            Window.findFrames(frames, test, true, isRedirected());
            
            frames.remove(this);
            
            if (frames.isEmpty())
               rects.add(test);
         }
      }
      return rects;
   }
   
   /**
    * Mark frame for special conditional down processing.
    * 
    * @param	mark 
    * 			<code>true</code> enables special down processing.
    */
   public void setHonorNextConditionalDown(boolean mark)
   {
      this.honorNextConditionalDown = mark;
   }

   /**
    * Check if frame requires special handling for conditional down.
    * 
    * @return   <code>true</code> if such a processing is required.
    */
   public boolean isHonorNextConditionalDown()
   {
      return honorNextConditionalDown;
   }

   /**
    * Get current state of fixed width flag.
    * 
    * @return  <code>true</code> if frame width is specified explicitly.
    */
   public boolean isFixedWidth()
   {
      return fixedWidth;
   }

   /**
    * Detects whether the frame was displayed with headers on the current
    * output page (of a paged stream) or ever (of a non-paged stream).
    * <p>
    * This determination makes sense for the down frames only.
    * <p>
    * It is based on a list of paged streams on which this frame has been
    * viewed for the current page. Each stream has its own paging state and
    * will change pages at a different/arbitrary point than any other stream.
    * <p>
    * This same list will also contain all non-paged streams which have ever
    * gotten output from this frame.
    *
    * @param    stream
    *           The stream to lookup in the list.
    * 
    * @return   <code>true</code> if the frame has been viewed for the current
    *           page on the given paged stream OR if the frame has been viewed
    *           at least once for a non-paged stream.  <code>false</code> if
    *           the frame is not a down frame or if it hasn't been viewed as
    *           noted above for the given stream.
    */
   public boolean isWithHeaders(Stream stream)
   {
      if (!down)
         return false;

      return headersOutput != null && headersOutput.contains(stream);
   }

   /**
    * Checks the touched state of the current iteration of the down frame.
    *
    * @return   <code>true</code> if the current iteration has received some
    *           output since the last cursor movement.
    */
   public boolean isCurrentTouched()
   {
      if (!down)
         return false;

      int cur = cursor < 0 ? 0 : cursor;
      return  touched[cur];
   }

   /**
    * Mark frame as boxed or non-boxed depending on NO-BOX flag.
    * 
    * @param   box
    *          <code>true</code> if frame is boxed and <code>false</code>
    *          otherwise.
    */
   public void setBox(boolean box)
   {
      frameScroll.setBorder(box ? new LineBorder() : new EmptyBorder());
   }
   
   // Container specific methods
   /**
    * Getting the content pane for scrollable container
    * 
    * @return  The Container handle for the peer container's content.
    */
   public ScrollContainer getContentPane()
   {
      return contentPane;
   }

   /**
    * Sets the custom layout to meet the Progress needs.
    * 
    * @param   customLayout
    *          The new layout to be used for widget whose coordinate is not
    *          explicitly defined upon adding to the peer frame.
    */
   @Override
   public void setLayout(LayoutManager customLayout)
   {
      if(customLayout instanceof ColumnLayout ||
         customLayout instanceof ZeroColumnLayout)    
      {
         getContentPane().setLayout(customLayout);          
      }
      super.setLayout(customLayout);
   }

   /**
    * Sets the size of the peer container. This is the visible size of the
    * container and can not be the same as the inner size if the container is
    * scrollable.
    * 
    * @param   width
    *          The new width of the container in character units.
    * @param   height
    *          The new height of the container in character units.
    */
   public void setSizeImpl(int width, int height)
   {
      if (contentPane != null && frameScroll != null)
      {
         int hdiff = calculateWidthDelta();
         int vdiff = calculateHeightDelta();
         
         contentPane.setVisibleSize(width - hdiff, height - vdiff);
         frameScroll.setSize(width - hdiff, height - vdiff);
      }

      try
      {
         // ignore redundant requests if config is being updated
         if (isConfigUpdateActive())
            return;
      
         Dimension screen = screen().screenDimension();

         screen = new Dimension(screen.width, 
                                screen.height - Window.RESERVED_LINES);

         int oldWidth  = width;
         int oldHeight = height;

         width  = adjustWidth(width, screen.width);
         height = adjustHeight(height, screen.height);

         setPaneVisibleSize(width, oldWidth - width, oldHeight - height);
         centerFrame(width, screen);
      }
      finally
      {
         if (config.getWidthChars() < width)
            config.setWidthCharsDirect(width);
         if (config.getHeightChars() < height)
            config.setHeightCharsDirect(height);

         setSizeInt(width, height);
      }
   }

   /**
    * Center frame at the screen, if necessary.
    * 
    * @param   width
    *          Frame width.
    * @param   screen
    *          Screen dimensions.
    */
   private void centerFrame(int width, Dimension screen)
   {
      // If the frame is centered adjust the location
      if (config.isCentered() && screen.width >= width)
      {
         setLocation((screen.width - width) / 2, location().y);
      }
   }

   /**
    * Calculate initial frame column.
    * 
    * @return  calculated column.
    */
   private int calculateColumn()
   {
      int col = config.getColumn();

      return col == BaseConfig.INV_COORD ? 0 : col;
   }

   /**
    * Set visible size of the pane.
    * 
    * @param   width
    *          Frame width.          
    * @param   deltaX
    *          Delta width.
    * @param   deltaY
    *          Delta height.
    */
   private void setPaneVisibleSize(int width, int deltaX, int deltaY)
   {
      if (contentPane != null && (deltaX != 0 || deltaY != 0))
      {
         int cols = config.getWidthChars();

         // explicitly set width is larger than screen
         if (cols > width)
         {
            config.setVirtualWidthCharsDirect(cols);
            contentPane.setSize(cols, 0);
         }

         Dimension d = contentPane.getVisibleSize();
         contentPane.setVisibleSize(d.width - deltaX, d.height - deltaY);
         frameScroll.minimumSize();
      }
   }

   /**
    * Adjust frame height.
    * 
    * @param   height
    *          Input height.
    * @param   screenHeight
    *          Screen height. 
    *          
    * @return  adjusted height.
    */
   private int adjustHeight(int height, int screenHeight)
   {
      if (config.getRow() != BaseConfig.INV_COORD     &&
          (config.getRow() + height) >  screenHeight &&
          config.getRow() < screenHeight)
      {
         height = screenHeight - config.getRow();

         if (height < 1)
            height = 1;
      }
      return height;
   }

   /**
    * Adjust frame width.
    * 
    * @param   width
    *          Input width.
    * @param   screenWidth
    *          Screen height. 
    *          
    * @return  adjusted height.
    */
   private int adjustWidth(int width, int screenWidth)
   {
      // TODO: note that in case if column is outside visible area, we do 
      // nothing but it is not clear yet if Progress does the same
      
      int col = calculateColumn();

      if (!config.isCentered()        && 
          (col + width) > screenWidth &&
          col < screenWidth)
      {
         width = screenWidth - col;
      }
      return width;
   }

   /**
    * Setting the inner size of the container.
    * 
    * @param   innerWidth
    *          The new inner width of the container in character units.
    * @param   innerHeight
    *          The new inner height of the container in character units.
    */
   public void setInnerSize(int innerWidth, int innerHeight)
   {
      if (contentPane != null)
         contentPane.setSize(innerWidth, innerHeight);

      config.setVirtualHeightCharsDirect(innerHeight);
      config.setVirtualWidthCharsDirect(innerWidth);
   }

   /**
    * Getting the inner size of the container. Can be more than the visible
    * size for scrollable containers.
    * 
    * @return  Dimension object which contains the inner size of the container.
    */
   public Dimension getInnerSize()
   {
      if (contentPane != null)
         return contentPane.dimension();

      return null;
   }

   /**
    * Moving the input focus to the first widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the last one within the container and
    * we have to simulate the Progress behavior in which the next focused
    * widget will be the first widget which can receive a focus in a
    * container.
    */
   @Override
   public void nextFocus()
   {
      Widget[] c = getContentPane().widgets();

      for (int i = 0; i < c.length; i++)
      {
         if (c[i].focusTraversable())
         {
            c[i].requestFocus();
            break;
         }
      }
   }
   
   /**
    * Moving the input focus to the last widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the first one within the container and
    * we have to simulate the Progress behavior in which the previous focused
    * widget will be the last widget which can receive a focus in a container.
    */
   public void previousFocus()
   {
      Widget[] c = getContentPane().widgets();

      for (int i = c.length - 1; i >= 0; i--)
      {
         if (c[i].focusTraversable())
         {
            c[i].requestFocus();
            break;
         }
      }
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per frame.
    * 
    * @param   origin
    *          The coordinate of the upper-left corner for peer container in
    *          character units.
    * @param   size
    *          The size of the peer container in character units.
    */
   public abstract void draw(Point origin, Dimension size);

   /**
    * Draw all uninitialized widgets of the current iteration.
    */
   public abstract void drawUninitialized();
   
   /**
    * Notify the registered object that the given stream has just written
    * a page break (and thus a new page of output has just begun).
    * <p>
    * This notification will only be provided for the second and subsequent
    * pages since it is based on notifying for page breaks (and there is no
    * page break for the first page).
    * <p>
    * This particular listener uses the page break notification as a signal
    * that the next time the frame outputs something, the frame headers should
    * be redrawn, since it will be the first occurrence of the frame on the new
    * page. Two conditions should be met for redrawing the frame headers:
    * <ul>
    *  <li>the stream should not be in the headers output set
    *  <li>the frame should be marked as hidden
    * </ul>
    * The very next drawing would make the frame visible, draw the whole frame
    * and insert the stream into the headersOutput set to avoid multiple drawings
    * of the frame headers on the same page.
    * The same logic applies when multiple frames output to the stream
    * alternatingly.
    *
    * @param    out
    *           The stream that just wrote a page break.
    */
   @Override
   public void notifyPageBreak(Stream out)
   {
      boolean redir = isRedirected();

      if (headersOutput != null && !out.isSuppressFormFeed())
      {
         headersOutput.remove(out);
      }
      
      // do not hide frame if it is not redirected
      if (redir)
         forceVisibility(false);
   }
   
   /**
    * Notify the registered object that the given stream is being closed.
    * <p>
    * Not used here.
    *
    * @param    out
    *           The stream being closed.
    */
   @Override
   public void notifyClose(Stream out)
   {
      // do nothing
   }
   
   /**
    * Notify registered object that given stream reached end of life cycle and
    * all references to it can be safely discarded.
    * <p>
    * This is used to remove non-paged streams from the list of streams for
    * which we have output headers already. Paged streams always get a
    * page break notification (even on the last page so long as at least 1
    * byte has already been output, which it must have in our case since it
    * is in the list). But non-paged streams don't get any other notification
    * that we can use for removal.
    * 
    * @param    out
    *           The stream being closed.
    */
   @Override
   public void notifyUnbind(Stream out)
   {
      if (headersOutput != null)
      {
         headersOutput.remove(out);
      }
   }
   
   /**
    * Get current state of labels placement flag.
    * 
    * @return  Returns the state of the flag.
    */
   public boolean isSideLabels()
   {
      return contentPane.isSideLabels();
   }
   
   /**
    * Set new state of the labels placement flag.
    * 
    * @param   sideLabels
    *          New state of the flag.
    */
   public void setSideLabels(boolean sideLabels)
   {
      contentPane.setSideLabels(sideLabels);
   }

   /**
    * Get current state of labels flag.
    * 
    * @return  Returns the state of the flag.
    */
   public boolean isLabels()
   {
      return contentPane.isLabels();
   }
   
   /**
    * Set new state of the labels placement flag.
    * 
    * @param   labels
    *          New state of the flag.
    */
   public void setLabels(boolean labels)
   {
      contentPane.setLabels(labels);
   }

   /**
    * Get current state of the ATTR-SPACE attribute.
    * 
    * @return  Returns the value of the attribute.
    */
   public boolean isAttrSpace()
   {
      return contentPane.isAttrSpace();
   }
   
   /**
    * Set new state of the ATTR-SPACE attribute.
    * 
    * @param   attr_space
    *          New state of the flag.
    */
   public void setAttrSpace(boolean attr_space)
   {
      contentPane.setAttrSpace(attr_space);
   }

   /**
    * Sets the number of columns (n COLUMNS) option.
    *
    * @param columns
    *        number of columns in the frame
    */
   public void setColumns(int columns)
   {
      if (columns < 0)
         columns = 0;
      
      LayoutManager manager = columns == 0 ? new ZeroColumnLayout()
                                           : new ColumnLayout(columns);
      setLayout(manager);
   }

   /**
    * Retrieve the value of the component at the specified row.
    * 
    * @param    index
    *           The row from which to retrieve data.
    * @param    comp
    *           The component which forms a column of values.
    *          
    * @return   The data from the specified location or "" if the frame is
    *           not a down frame or if the location contained no data.
    */
   public String getFrameValue(int index, Widget comp)
   {
      if (!down)
      {
         return "";
      }
      
      Widget field = getField(index, comp);
      
      if (!(field instanceof DataContainer))
      {
         return "";
      }
      
      return getFieldText(field);
   }

   /**
    * Get text of field.
    * 
    * @param   field
    *          Field to get text from.
    *          
    * @return  field text.
    */
   private String getFieldText(Widget field)
   {
      String text = "";
      
      if (field.<WidgetConfig>config().getState() != ScreenBuffer.UNINITIALIZED)
      {
         text = ((DataContainer) field).getText();
      }
      
      if (text == null || text.trim().length() == 0)
      {
         text = "";
      }
      
      return text;
   }
   
   /**
    * Get vertical size of fixed portion of pane (like headers and such).
    * 
    * @return  The height of fixed portion of pane.
    */
   public int getFixedHeight()
   {
      return headersHeight;
   }
   
   /**
    * Get vertical size of one iteration.
    * 
    * @return  The iteration height.
    */
   public int getIterationHeight()
   {
      return verticalStep;
   }

   /**
    * Get size of the leading skip in iteration.
    * 
    * @return   size of the leading skip.
    */
   public int getLeadingSkipSize()
   {
      return leadingSkipSize ;
   }
   
   /**
    * Get size of the trailing skip in iteration.
    * 
    * @return   size of the trailing skip.
    */
   public int getTrailingSkipSize()
   {
      return trailingSkipSize;
   }
   
   /**
    * Get field from specified iteration.
    * 
    * @param   index
    *          Row to retrieve.
    * @param   comp
    *          Reference component.
    *          
    * @return  Located component.
    */
   public Widget getField(int index, Widget comp)
   {
      if (!down || downBody == null || comp == null)
         return comp;
      
      WidgetConfig cfg = comp.config();
      
      for (int i = 0; i < downBody[index].length; i++)
      {
         WidgetConfig tmp = downBody[index][i].config();
         
         if (tmp.getID() == cfg.getID())
            return downBody[index][i];
      }
      
      // return source field if nothing is found in the down frame body
      return comp;
   }
   
   /**
    * Clear all fill-ins in frame and mark all iterations in down frame as
    * untouched.
    * 
    * @param   clearAll
    *          Force clearing all widgets in down frame.
    */
   public void clearWorker(boolean clearAll)
   {
      Widget[] list = getContentPane().widgets();

      if (!down)
      {
         clearComponents(list);
      }
      else
      {
         if (downBody == null)
            return;
         
         resetDownBody(0, downBody.length, clearAll);
      }
   }

   /**
    * Reset widgets which belong to DOWN body.
    * 
    * @param   from
    *          Start row.
    * @param   to
    *          End row.
    * @param   clearAll
    *          if <code>true</code> then all widgets will be cleared regardless
    *          from the row clear flag. 
    */
   private void resetDownBody(int from, int to, boolean clearAll)
   {
      for (int i = from; i < to; i++)
      {
         underlinedRow[i] = false;
         
         if (!needClear[i] && !clearAll)
            continue;
         
         resetDownBodyWorker(i);
      }
      readjustDownBody();
   }

   /**
    * Worker routine for {@link #resetDownBody(int, int, boolean)}.
    * 
    * @param   idx
    *          Row index.
    */
   private void resetDownBodyWorker(int idx)
   {
      needClear[idx] = false;
      touched[idx] = false;
      
      for (int j = 0; j < downBody[idx].length; j++)
      {
         underline[idx][j] = false;
         
         if (!UiUtils.hasConfig(downBody[idx][j]))
            continue;
         
         ComponentConfig cc = downBody[idx][j].config();
         
         if ((cc instanceof ControlTextConfig) && ((ControlTextConfig)cc).isStatic())
            continue;
         
         cc.setState(ScreenBuffer.UNINITIALIZED);
         
         if (idx > 0 && downBody[idx][j].isEnabled())
         {
            downBody[0][j].setEnabled(true);
            downBody[idx][j].setEnabled(false);
         }
         
         // reset colors
         if (cc instanceof BaseConfig)
         {
            cc.setDcolor(Color.NORMAL);
            cc.setPfcolor(Color.NORMAL);
         }
      }
   }

   /**
    * Reset widgets in regular frame.
    */
   public void resetAll()
   {
      if (down)
         return;
      
      for (Widget comp : getContentPane().widgets())
      {
         if (comp instanceof FillIn)
         {
            ((FillIn) comp).reset();
         }
      }
   }
   
   /**
    * Clears all fill-ins of the current iteration.
    */
   public void clearCurrent()
   {
      Widget[] list = getContentPane().widgets();

      if (!down)
      {
         clearComponents(list);
      }
      else
      {
         if (downBody == null)
            return;
         
         int from = cursor < 0 ? 0 : cursor;

         resetDownBody(from, from + 1, true);
      }
   }

   /**
    * Repaints the current iteration.
    */
   public void repaintCurrent()
   {
      if (!down || downBody == null)
      {
         return;
      }

      int ix = cursor < 0 ? 0 : cursor;

      for (int i = 0; i < downBody[ix].length; i++)
      {
         Widget item = downBody[ix][i];

         item.repaint();

         // repaint side labels if any
         if (isSideLabels())
         {
            if (item instanceof LabeledWidget)
            {
               Label label = ((LabeledWidget) item).getLabelInstance();

               if (label != null)
                  label.repaint();
            }
         }
      }
   }

   /**
    * Get IDs of widgets which belong to single iteration.
    * 
    * @return  {@link Set} of widget IDs.
    */
   public Set<Integer> getIterationIds()
   {
      if (!down || downBody == null)
         return null;
      
      Set<Integer> res = new HashSet<Integer>();
      
      for (int i = 0; i < downBody[0].length; i++)
      {
         Widget item = downBody[0][i];
         
         if (UiUtils.hasConfig(item))
         {
            int id = item.<WidgetConfig>config().getID();
            
            res.add(Integer.valueOf(id));
         }
      }
      
      return res;
   }
   
   /**
    * Marks the current iteration as touched to allow all uninitialized
    * widgets to display. For redirected frames with trailing non-empty SKIPs
    * it also makes sure that clipping rectangle is set correctly.
    */
   public void touchRow()
   {
      if (!down)
         return;
      
      touched[cursor] = true;

      if (!isVisible() || !isRedirected())
         return;
      
      for (int i = downBody[cursor].length - 1; i >= 0; i--)
         downBody[cursor][i].repaint();
   }
   
   /**
    * Mark current iteration as such that need to be cleared.
    */
   public void needClearRow()
   {
      if (!down)
         return;
      
      int row = (cursor < 0) ? 0 : cursor; 
      
      if (needClear == null)
         pendingNeedClear  = true;
      else
         needClear[row] = true;
   }
   
   /**
    * Reset CHANGED flag in fill-in widgets.
    */
   public void resetChanged()
   {
      
      if (!down)
      {
         Widget[] list = getContentPane().widgets();
         for (int i = 0; i < list.length; i++)
         {
            Widget widget = list[i];
            
            if (UiUtils.hasConfig(widget))
               resetChangedFlag(widget);
         }
      }
      else
      {
         for (int i = 0; i < downBody.length; i++)
         {
            for (int j = 0; j < downBody[i].length; j++)
            {
               Widget widget = downBody[i][j];
               
               if (widget instanceof FillIn)
                  resetChangedFlag(widget);
            }
         }
         
      }
   }

   /**
    * Reset {@link ScreenBuffer#CHANGED} flag for specified widget.
    * 
    * @param   widget
    *          Input widget.
    */
   private void resetChangedFlag(Widget widget)
   {
      ComponentConfig cfg = widget.config();
      
      if (cfg.getState() == ScreenBuffer.CHANGED)
      {
         cfg.setState(ScreenBuffer.UNCHANGED);
      }
   }

   /**
    * Set frame value.
    * 
    * @param    comp
    *           Component which is used as a reference.
    * @param    value
    *           Value to set.
    */
   public void setValue(Widget comp, BaseDataType value)
   {
      setValue(comp, value, null);
   }
      
   /**
    * Set frame value.
    * 
    * @param    comp
    *           Component which is used as a reference.
    * @param    value
    *           Value to set.
    * @param    fmt
    *           An optional format string to use with this value.  This will
    *           temporarily override the current format string if and only if
    *           <code>component</code> is a <code>FillIn</code>. May be 
    *           <code>null</code>.
    */
   public void setValue(Widget comp, BaseDataType value, String fmt)
   {
      if (down)
      {
         // dereference the proper iteration
         comp = getField(comp);
      }
      
      if (comp instanceof DataContainer)
      {
         if (comp instanceof FillIn)
         {
            ((FillIn) comp).setValue(value, fmt);
         }
         else
         {
            ((DataContainer) comp).setValue(value);
         }
         
         comp.<WidgetConfig>config().setState(ScreenBuffer.UNCHANGED);
      }
   }

   /**
    * Set field state.
    * 
    * @param   comp
    *          Reference component.
    * @param   state
    *          New state for the field.
    */
   public void setState(Widget comp, int state)
   {
      if (down)
      {
         // down frame
         Widget field = getField(comp);
         
         if (UiUtils.hasConfig(field))
         {
            field.<WidgetConfig>config().setState(state);
            return;
         }
      }
      
      // regular frame or component is not part of the down frame body
      comp.<WidgetConfig>config().setState(state);
   }
   
   /**
    * Reset widget when switching from UNINITIALIZED state.
    * 
    * @param   comp
    *          Reference component.
    */
   public void resetWidget(Widget comp)
   {
      Widget field = getField(comp);
      
      if (comp instanceof RadioSet)
         ((RadioSet) field).reset();
   }
   
   /**
    * Apply config to specified widget.
    * 
    * @param   comp
    *          Reference component.
    * @param   cfg
    *          Config to apply.
    */
   public void applyConfig(Widget comp, ComponentConfig cfg)
   {
      Widget field = getField(comp);
      
      if (!down || field == null)
      {
         // apply to base component but preserve state
         ComponentConfig config = comp.config();
         int state = config.getState();
         
         boolean visible = isVisible();
         boolean newVis  = cfg.isVisible(); 
         boolean canHide = isCanHide();
         int down = getDown();
         
         if (comp == this)
         {
            ((FrameConfig) config).setDown(down);
            cfg.setVisibleDirect(visible);
         }
         
         Color dcolor = null;
         Color pfcolor = null;
         
         if (config instanceof BaseConfig)
         {
            dcolor  = ((BaseConfig) config).getDcolor();
            pfcolor = ((BaseConfig) config).getPfcolor();
         }
         
         config.applyConfig(cfg);
         config.setState(state);
         
         // restore state of peer frame
         if (comp == this)
         {
            if(visible)
            {
               // just restore flag, frame visibility remains
               // unchanged so there is no need to restore it
               setCanHide(canHide);
            }
            else if (!visible && newVis) 
            {
               // frame visibility is changed
               postponeSetVisible(true);
               postponePlace(true);
            }
         }
         else
         {
            if (config instanceof BaseConfig)
            {
               BaseConfig bcfg = (BaseConfig) config;

               if (!Utils.equals(dcolor, bcfg.getDcolor()) ||
                   !Utils.equals(pfcolor, bcfg.getPfcolor()))
               {
                  colorChanged();

                  if (isVisible() && field != null)
                     field.repaint();
               }
            }
         }

         return;
      }

      // down frame
      if (field.config() == null)
         return;

      ComponentConfig config = comp.config();
      
      // if the frame was moved, we must use the position received from the
      // server-side
      boolean frameMoved = comp instanceof Frame && 
                           ((FrameConfig) cfg).isPositionChanged();
      
      if (frameMoved == false)
      {
         // fill some values to preserve layout 
         Point loc = field.location();
         
         cfg.setRow(loc.y);
         
         // change only COLUMN (i.e. left aligned widgets), 
         // but don't touch COLON and TO
         if (cfg instanceof BaseConfig)
         {
            BaseConfig bcfg = (BaseConfig) cfg;
            
            if (bcfg.isLeftAligned())
               bcfg.setColumn(loc.x);
         }
         else
            cfg.setColumn(loc.x);
      }
      else
      {
         // let local copy of config know that this is application triggered 
         // frame move
         ((FrameConfig) config).setPositionChanged(true);
      }

      // apply label, if any, to base component
      String label = cfg.getLabel();
      
      config.setLabel(label);
      
      // preserve state of the labeling flags if filling not first row
      if (field instanceof LabeledWidget &&
          comp != field)
      {
         LabeledWidget lw = (LabeledWidget) field;

         if (lw.isNoLabels())
            cfg.setLabels(false);
      }

      // finally apply config but preserve current state
      config = field.config();
      int state = config.getState();
      
      Color dcolor = null;
      Color pfcolor = null;
      
      if (config instanceof BaseConfig)
      {
         dcolor  = ((BaseConfig) config).getDcolor();
         pfcolor = ((BaseConfig) config).getPfcolor();
      }
      
      config.applyConfig(cfg);
      config.setState(state);
      
      if (frameMoved)
         ((FrameConfig) config).setPositionChanged(false);

      if (config instanceof BaseConfig)
      {
         BaseConfig bcfg = (BaseConfig) config;
         
         if (!Utils.equals(dcolor, bcfg.getDcolor()))
         {
            colorChanged();
         }
         else
         {
            if (!Utils.equals(pfcolor, bcfg.getPfcolor()))
            {
               if (field.hasFocus())
                  colorChanged();
            }
            else if (field instanceof FillIn)
            {
               if (((FillIn) field).isSelected())
                  colorChanged();
            }
         }
      }
   }

   /**
    * Retrieve text values for entire column.
    * 
    * @param   focusedCol
    *          Reference widget.
    *          
    * @return  Array of text values.
    */
   public String[] getValues(Widget focusedCol)
   {
      if (!down)
         return null;
      
      String[] res = new String[config.getDown()];
      
      for (int i = 0; i < res.length; i++)
      {
         Widget comp = getField(i, focusedCol);
         
         if (comp instanceof DataContainer)
            res[i] = (((DataContainer) comp).getText()).trim();
         else
            res[i] = "";
      }
      return res;
   }

   /**
    * Set visibility of the down frame.
    * 
    * @param   visible
    *          New visibility state.
    */
   public void setBodyVisible(boolean visible)
   {
      contentPane.setVisible(visible);
      
      if (downBody != null)
      {
         for (int i = 1; i < downBody.length; i++)
         {
            for (int j = 0; j < downBody[i].length; j++)
            {
               downBody[i][j].setVisible(visible);
            }
         }
      }

      if (nonbody != null)
      {
         // adjust VISIBLE flag for non-body widgets too
         for (int i = 0; i < nonbody.length; i++)
         {
            ThinClient.getInstance().collectFlag(nonbody[i], visible, true);
            nonbody[i].setVisible(visible);
         }
      }
   }
   
   /**
    * Return <code>true</code> if frame is pending fixed down.
    *
    * @return <code>true</code> if frame is pending fixed down.
    */
   public boolean isPendingFixedDown()
   {
      return pendingFixedDown;
   }

   /**
    * Return the HEIGHT-CHARS Progress attribute that was specified upon frame 
    * creation.
    *
    * @return see above.
    */
   public int getOriginalHeightChars()
   {
      return originalHeightChars;
   }

   /**
    * Set the pendingFixedDown flag.
    *
    * @param   state
    *          New state for this flag.
    */
   public void setPendingFixedDown(boolean state)
   {
      pendingFixedDown = state;
   }

   /**
    * Return <code>true</code> if frame is down frame.
    * 
    * @return <code>true</code> if frame is down frame.
    */
   public boolean isDown()
   {
      return down;
   }
   
   /**
    * Set the down state of this frame.
    *
    * @param   down
    *          The down-state of this frame.
    */
   public void setDown(boolean down)
   {
      Frame.this.down = down;
   }
   
   /**
    * Underlines a set of widgets.
    * 
    * @param   list
    *          array of field level widgets to be underlined.
    */
   public void underline(Widget[] list)
   {
      if (!down)
         return;

      if (cursor < 0)
         cursor = 0;
      
      for (int i = 0; i < list.length; i ++)
      {
         int ndx = getFieldIndex(list[i]);
         
         if (ndx < 0)
         {
            throw new 
               RuntimeException("Invalid component requested for UNDERLINE: " 
                                + list[i]);
         }
         
         underline[cursor][ndx] = true;
      }
      
      underlinedRow[cursor] = true;
      touched[cursor] = true;
      
      readjustDownBody();
      repaint();
   }

   /**
    * Check if down frame elements size calculation is performed at least
    * once.
    * 
    * @return  <code>true</code> if down frame elements are calculated.
    */
   public boolean isDownCalculated()
   {
      return saveDown >= 0;
   }
   
   /**
    * Fixes the list of non-body components of this down frame in case there
    * is a replacement.
    * 
    * @param   oldc
    *          existing component to be replaced in the non-body list
    * @param   newc
    *          replacement component
    */
   public void fixNonBody(Widget oldc, Widget newc)
   {
      // nothing to do?
      if (!down || nonbody == null)
         return;
      
      // replacement
      for (int i = 0; i < nonbody.length; i ++)
         if (nonbody[i].equals(oldc))
            nonbody[i] = newc;
   }
   
   /**
    * Finish previous iteration and save frame buffer data into 'spare'
    * iteration.
    *
    * @param   widget
    *          The frame or widget.
    * @param   frameBuf
    *          The current values of widgets.
    * @param   widgetIds
    *          Array of field level widget IDs.
    * @param   isView
    *          <code>true</code> if parameters are for VIEW,
    *          <code>false</code> for DISPLAY.
    */
   public void prepareIteration(Widget widget,
                         ScreenBuffer frameBuf,
                         int[] widgetIds,
                         boolean isView)
   {
      if (pendingCursorWrap)
         finishCursorWrap(true, true);

      spareIteration = new SpareIteration(widget, frameBuf, widgetIds, isView);
   }
   
   /**
    * Get field from current iteration.
    * 
    * @param   comp
    *          Reference component.
    *          
    * @return  Located component.
    */
   public Widget getField(Widget comp)
   {
      int row = (cursor < 0) ? 0 : cursor; 
      
      return getField(row, comp);
   }
   
   /**
    * Scans the list of widgets and detects the case when there is a mixture
    * of regular widgets with one or more trailing skips.
    */
   void detectTrailingSkips()
   {
      Widget[] c = contentPane.widgets();
      boolean metSkip = false;
      
      // the loop below scans the array of frame widgets backwards.
      // the first encountered regular widget breaks the loop and sets
      // the mixed flag to true if during the scan the skips were met.
      for (int i = c.length - 1; i >= 0; i--)
      {
         if ((c[i] instanceof Skip))
         {
            metSkip = true;
            continue;
         }
         
         trailingSkips = metSkip;
         return;
      }
   }

   /**
    * Check if the specified widget is the last component in this frame.
    * <p>
    * Note that last widget is determined as last traversable widget in Tab
    * order, regardless from its position in the frame. Since
    * {@link ThinClient#enable(int, int[], ScreenBuffer, boolean)}
    * changes Tab order, it also changes which widget is last by the means
    * of this method.
    * 
    * @param   wdgt
    *          widget to check
    * 
    * @return  <code>true</code> if this component is the last component in
    *          the frame.
    */
   public boolean isLastWidget(Widget wdgt)
   {
      Widget[] list = getContentPane().widgets();
      
      for (int i = (list.length - 1); i >= 0; i--)
      {
         if (list[i].focusTraversable())
            return (list[i] == wdgt) ? true : false;
      }
      
      return false;
   }

   /**
    * Clears the pending "moved cursor" condition and puts the iterations
    * widgets into their visible cursor position.
    */
   void viewed()
   {
      if (!down)
         return;
   }

   /**
    * Moves the down frame cursor up or down by a specified number of
    * iterations.
    * <p>
    * Moving up beyond the top iteration means wrapping to the bottom. Moving
    * down beyond the last iteration means wrapping back to the top. In both
    * cases the down frame gets cleared.
    * 
    * @param    by
    *           number of lines; positive numbers mean moving cursor down
    * @param    noPause
    *           if <code>true</code> then no pause will be generated on cursor
    *           wrap.
    * 
    * @return   <code>true</code> if cursor wrap has occurred
    */
   public boolean moveCursor(int by, boolean noPause)
   {
      // check conditions
      if (!down)
         return false;
      
      if (cursor < 0)
         cursor = 0;
      
      // calculate new position
      
      pendingCursorPos = cursor + by;

      boolean needPause = false;

      if (pendingCursorPos < 0)
      {
         needPause = true;
         pendingCursorPos = lastRow;
         newRow = -2;
      }

      if (pendingCursorPos > lastRow)
      {
         pendingCursorPos = 0;

         // for non-down frames, if we go past the last row during a display
         // or a view when the frame is hidden, we go back to row 0 without
         // clearing the frame (i.e. no pause is needed)
         if (isDown() && isVisible() == false)
         {
            needPause = false;
         }
         else
         {
            needPause = true;
            newRow = lastRow + 1;
         }

          clearMode = !isVisible();
      }
      
      boolean normalExit = false;

      if (needPause)
      {
         ThinClient.wrapFrame(this);
         
         pendingCursorWrap = true;

         try
         {
            // If pausing is enabled, then process pausing and clearing 
            // immediately, otherwise simulate abnormal exit which will 
            // trigger frame repainting on next output to frame.
            if (!noPause)
            {
               if (isNeedPause())
               {
                  if (isVisible())
                     ThinClient.getInstance().pause(-1, null, false);
               }
               else
                  setNeedPause(true);

               normalExit = true;
            }
         }
         finally
         {
            finishCursorWrap(normalExit, false);
         }

         cursor = pendingCursorPos;
      }
      else
      {
         cursor = pendingCursorPos;

         if (clearMode)
            clearCurrent();
      }

      return needPause;
   }

   /**
    * Scrolls the down frame up or down by one line. 
    * 
    * @param   fromCurrent
    *          <code>true</code> if FROM-CURRENT option was used
    * @param   upDown
    *          <code>true</code> if scrolling up
    */
   public void scroll(boolean fromCurrent, boolean upDown)
   {
      if (!down)
         return;
      
      if (cursor < 0)
         cursor = 0;
      
      int maxDown = config.getDown() - 1;
      int current = (fromCurrent) ? cursor : 0;
      
      if (upDown)
      {
         scrollWorker(current + 1, maxDown, current, 1, maxDown);
      }
      else
      {
         scrollWorker(maxDown - 1, current, maxDown, -1, current);
      }
   }

   /**
    * Down frame body scrolling worker routine.
    * 
    * @param   from
    *          Source start row.
    * @param   to
    *          Source end row.
    * @param   moveTo
    *          Destination start row.
    * @param   incr
    *          Increment.
    * @param   insertAt
    *          Row where empty row should be inserted.
    */
   private void scrollWorker(int from, int to, int moveTo, int incr,
                          int insertAt)
   {
      scrollDownBody(from, to, moveTo, incr);
      insertEmptyRow(insertAt);
   }

   /**
    * Insert empty row at specified index.
    * 
    * @param   insertAt
    *          Index at which empty row should be inserted. 
    */
   private void insertEmptyRow(int insertAt)
   {
      // insert an empty line
      for (int j = 0; j < downBody[insertAt].length; j++)
      {
         if (downBody[insertAt][j] instanceof FillIn)
         {
            ((FillIn) downBody[insertAt][j]).clear();
         }
         
         ComponentConfig nc = downBody[insertAt][j].config();
         
         nc.setState(ScreenBuffer.UNINITIALIZED);
      }

      // insert underlinedRow[] and touched[] flags
      underlinedRow[insertAt] = false;
      touched[insertAt] = false;
   }

   /**
    * Scroll down body.
    * 
    * @param   from
    *          Source start row.
    * @param   to
    *          Source end row.
    * @param   moveTo
    *          Destination start row.
    * @param   incr
    *          Increment.
    */
   private void scrollDownBody(int from, int to, int moveTo, int incr)
   {
      // move lines
      for (int i = from;
          (incr > 0 && i <= to) || (incr < 0 && i >= to); 
           i += incr, moveTo += incr)
      {
         //lines[moveTo] = lines[i];
         for (int j = 0; j < downBody[i].length; j++)
         {
            ComponentConfig oc = downBody[i][j].config();
            ComponentConfig nc = downBody[moveTo][j].config();
            nc.setState(oc.getState());
            
            if (!(downBody[i][j] instanceof DataContainer) || 
                oc.getState() == ScreenBuffer.UNINITIALIZED)
            {
               if (downBody[i][j] instanceof DataContainer)
                  downBody[moveTo][j].repaint();
               continue;
            }
                        
            BaseDataType val = ((DataContainer) downBody[i][j]).getValue();
            ((DataContainer) downBody[moveTo][j]).setValue(val);
         }
         
         // scroll underlinedRow[] and touched[] flags
         underlinedRow[moveTo] = underlinedRow[i];
         touched[moveTo] = touched[i];         
      }
   }

   /**
    * Clears the down frame and resets the cursor to the top of the frame.
    */
   public void clear()
   {
      if (!down)
         return;

      cursor = 0;
   }
   
   /**
    * Reset cursor to initial state.
    */
   void resetCursor()
   {
      if (!down)
         return;
      
      cursor = -1;
   }
   
   /**
    * Returns the current cursor position in this down frame.
    * 
    * @return   The current cursor position or 0 if this is not a down frame.
    */
   public int getCursor()
   {
      if (!down)
         return 0;
      
      return (pendingCursorWrap) ? newRow  : cursor;
   }
   
   /**
    * Assigns the current cursor position in a down frame. This is a NOP if
    * the frame is not a down frame.
    * 
    * @param    cursor
    *           New cursor position.
    */
   public void setCursor(int cursor)
   {
      if (!down)
         return;
      
      Frame.this.cursor = cursor;
   }
   
   /**
    * Fill down frame body.
    */
   void fillDownBody()
   {
      ArrayList<Widget> onerow = new ArrayList<Widget>(); 
      ArrayList<Widget> others = new ArrayList<Widget>();

      int ndown = config.getDown();
      
      if (!down || ndown == 0)
      {
         // make sure that call will be made only once
         if ((pendingFixedDown || (ndown == 1 && !down)) && verticalStep == 0)
         {
            calcFixedRect(onerow, others);
            nonbody = others.toArray(new Widget[others.size()]);
            if (ndown > 0)
            {
               downBody  = new Widget[ndown][];
               downBody[0] = onerow.toArray(new Widget[onerow.size()]);
            }
         }

         return;
      }

      if (saveDown == ndown && downBody != null)
      {
         readjustDownBody();
         return;
      }

      saveDown = ndown;
      
      calcFixedRect(onerow, others);
      
      nonbody = others.toArray(new Widget[others.size()]);
      
      boolean[][] savedUnderline = underline;
      boolean[] savedUnderlinedRow = underlinedRow;

      allocate(ndown);
      fillDownBodyParts(onerow, ndown);
      readjustDownBody();
      adjustFlagsAndCursor(savedUnderline, savedUnderlinedRow);
   }

   /**
    * Adjust cursor and flags after down frame body filling. 
    * 
    * @param   savedUnderline 
    *          Saved underline flag array.
    * @param   savedUnderlinedRow
    *          Saved underlined row array. 
    */
   private void adjustFlagsAndCursor(boolean[][] savedUnderline, 
                                     boolean[] savedUnderlinedRow)
   {
      if (pendingNeedClear)
      {
         needClear[0] = true;
         pendingNeedClear = false;
      }
      
      if (isRedirected())
      {
         int underlineCounter = Math.min(savedUnderline[0].length,
                                         underline[0].length);
         int underlinedRowCounter = Math.min(savedUnderlinedRow.length,
                                             underlinedRow.length);
         for (int i = 0; i < underlineCounter; i++)
            underline[0][i] = savedUnderline[0][i];
         for (int i = 0; i < underlinedRowCounter; i++)
            underlinedRow[i] = savedUnderlinedRow[i];
      }
            
      // adjust cursor position if it is outside new DOWN value 
      if (cursor > lastRow)
      {
         resetCursor();
         pendingCursorWrap = false;
      }
   }

   /**
    * Fill down body and relevant flags.
    * 
    * @param   onerow
    *          List of widgets consisting one row.
    * @param   ndown
    *          Number of iterations.
    */
   private void fillDownBodyParts(ArrayList<Widget> onerow, int ndown)
   {
      // First row consist of real widgets.
      downBody[0] = onerow.toArray(new Widget[onerow.size()]);
      
      detectSkips(downBody[0]);
      
      underline[0] = new boolean[downBody[0].length];
      
      // Remaining rows are artificial.
      for (int i = 1; i < ndown; i++)
      {
         underline[i] = new boolean[underline[0].length];
         needClear[i] = false;
         touched[i]   = false;
         
         ArrayList<Widget> row = new ArrayList<Widget>();
         
         for (Widget comp : onerow)
         {
            Widget copy = copyComponent(comp);
            contentPane.add(copy);
            row.add(copy);
         }            

         downBody[i] = row.toArray(new Widget[row.size()]);
      }
   }

   /**
    * Allocate down frame body parts.
    * 
    * @param   ndown
    *          Number of iterations.
    */
   private void allocate(int ndown)
   {
      downBody  = new Widget[ndown][];
      touched   = new boolean[ndown];
      underline = new boolean[ndown][];
      underlinedRow = new boolean[ndown];
      needClear = new boolean[ndown];
      startRow  = new int[ndown];
   }

   /**
    * Detect leading and trailing SKIP widgets in iteration.
    * 
    * @param    c
    *           Array of iteration widgets.
    */
   private void detectSkips(Widget[] c)
   {
      boolean hasOthers = false;
      
      leadingSkipSize  = 0;
      trailingSkipSize = 0;
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i].config() == null)
            continue;

         if (c[i] instanceof Skip)
         {
            if (!hasOthers)
               leadingSkipSize += c[i].height();
            else
               trailingSkipSize += c[i].height();
         }
         else
         {
            hasOthers = true;
            trailingSkipSize = 0;
         }
      }
      
      if (!hasOthers)
         leadingSkipSize = -1;
   }

   /**
    * Determine if the <code>WIDTH-CHARS</code> attribute can be set to the 
    * specified value.
    * 
    * @param   width
    *          The width value to be tested.
    * 
    * @return  <code>true</code> if the specified width is valid.
    */
   public boolean canSetWidthChars(int width)
   {
      if (width == 0 && config.getWidthChars() != 0)
      {
         return false;
      }

      if (sizeErrorCondition())
      {
         int minWidth = getMinimumSize().width;
         if (minWidth == 0)
         {
            minWidth = config.isBox() ? 2 : 1;            
         }
         else
         {
            minWidth += config.isBox() ? 1 : 0;
         }
         
         if (width != 0       &&
             width < minWidth && 
             isRedirected() == false)
         {
            displayError(4070, "Unable to set the WIDTH of the %s %s " +
                               "too small to contain all visible widgets");
            return false;
         }         
      }
      return true;
   }

   /**
    * Display error with specified code and message.
    * 
    * @param   code 
    *          Error code.
    * @param   err
    *          Error message.
    */
   private void displayError(int code, String err)
   {
      String type = config.isDialog() ? "DIALOG-BOX" : "FRAME";
      String msg = String.format(err, type, shortName());
      
      ErrorManager.displayError(code, msg);
   }

   /**
    * Determine if the <code>HEIGHT-CHARS</code> attribute can be set to the 
    * specified value.
    * 
    * @param   height
    *          The height value to be tested.
    * 
    * @return  <code>true</code> if the specified height is valid.
    */
   public boolean canSetHeightChars(int height)
   {
      if (sizeErrorCondition())
      {
         int minHeight = getFixedHeight();
         if (minHeight == 0)
         {
            minHeight = config.isBox() ? 2 : 1;
         }
         else
         {               
            minHeight += config.isBox() ? 1 : 0;
         }
         
         if (height != 0 && height < minHeight && isRedirected() == false)
         {
            displayError(4070, "Unable to set the HEIGHT of the %s %s " +
                               "too small to contain all visible widgets");
            return false;
         }         
      }
   
      return true;
   }

   /**
    * Check prerequisite conditions for potential size change errors.
    * 
    * @return  <code>true</code> if conditions are met.
    */
   private boolean sizeErrorCondition()
   {
      return isConfigUpdateActive() && config.isSizeChanged();
   }
   
   /**
    * Determine if the <code>ROW</code> attribute can be set to the specified 
    * value.
    * 
    * @param   row
    *          The row value to be tested.
    * 
    * @return  <code>true</code> if the specified row is valid.
    */
   public boolean canSetRow(int row)
   {
      if (posErrorCondition()         &&
          row < 0                     && 
          row != BaseConfig.INV_COORD &&
          isRedirected() == false)
      {
         displayError(4081, "Attribute ROW must be greater than or " +
                            "equal to 1 on %s %s");
         return false;
      }
      
      int height = config.getHeightChars();
      int screenHeight = Window.getScreenDimension().height;
      
      if (posErrorCondition()         &&
          row > screenHeight - height &&
          isRedirected() == false)
      {
         displayError(4054, "Unable to set ROW. %s %s does not fit " +
                            "in WINDOW widget");
         return false;
      }
   
      if (isRedirected())
      {
         // change the attribute value, but do not change the location
         config.setRowAttribute(row);
         return false;
      }
      
      return true;
   }

   /**
    * Determine if the <code>COLUMN</code> attribute can be set to the  
    * specified value.
    * 
    * @param   column
    *          The column value to be tested.
    * 
    * @return  <code>true</code> if the specified column is valid.
    */
   public boolean canSetColumn(int column)
   {
      if (!posErrorCondition()) 
         return true;
      
      if (column < 0 && column != BaseConfig.INV_COORD)
      {
         displayError(4081, "Attribute COL must be greater than or " +
                            "equal to 1 on %s %s");
         return false;
      }
   
      int width = config.getWidthChars();
      int screenWidth = Window.getScreenDimension().width;
      
      if (column > screenWidth - width)
      {
         displayError(4054, "Unable to set COL. %s %s does not fit " +
                            "in WINDOW widget");
         return false;
      }
      return true;
   }

   /**
    * Check prerequisite conditions for potential position change errors.
    * 
    * @return  <code>true</code> if conditions are met.
    */
   private boolean posErrorCondition()
   {
      return isConfigUpdateActive() && config.isPositionChanged();
   }

   /**
    * Clear specified components.
    * 
    * @param    list
    *           Components to clear.
    */
   private void clearComponents(Widget[] list)
   {
      for (Widget comp : list)
      {
         if (UiUtils.hasConfig(comp))
         {
            ComponentConfig cc = comp.config();
            
            if (!(cc instanceof ControlTextConfig) || 
                !((ControlTextConfig)cc).isStatic())
            {
               cc.setState(ScreenBuffer.UNINITIALIZED);

               // TODO: Probably this is more  correct than handling of 
               //       UNINITIALIZED in FillIn.draw(), but this may require
               //       much more testing and regression prone.
               /*
               if (comp instanceof FillIn)
                  ((FillIn) comp).setValue(chr);

               if (comp instanceof Editor)
                  ((Editor) comp).clear();
               */
            }
         }
      }
   }
   
   /**
    * Calculate height of fixed and repeated parts of frame.
    * 
    * @param   onerow
    *          List where widgets from repeated part of frame will be saved.
    * @param   others
    *          List where widgets from fixed part of frame will be saved.
    */
   private void calcFixedRect(ArrayList<Widget> onerow, 
                              ArrayList<Widget> others)
   {
      if (downBody != null)
      {
         for (int i = 1; i < downBody.length; i++)
            for (int j = 0; j < downBody[i].length; j++)
               contentPane.remove(downBody[i][j]);
      }
      
      Widget[] c = contentPane.widgets();
      Rectangle b = null;
      Rectangle b2 = null;
      
      for (int i = 0; i < c.length; i++)
      {
         if ((c[i] instanceof Label) ||
             (UiUtils.hasConfig(c[i]) && c[i].<WidgetConfig>config().isHeader()) ||
             (c[i] instanceof Space) ||
             (c[i].config() == null))
         {
            // all these widgets do not belong to down frame body
            
            if (c[i] instanceof Label && isSideLabels())
               continue;
       
            if (others != null)
               others.add(c[i]);
            continue;
         }
        
         if (!(c[i] instanceof Skip))
         {
            if (b == null)
               b2 = b = c[i].bounds();
            else
               b2 = b = b.union(c[i].bounds());
         }
         else
         {
            if(b == null)
               b = new Rectangle(c[i].location(),
                                 new Dimension(1, c[i].height()));
            else
               b = new Rectangle(b.top(), b.left(), 
                                 b.bottom() +  c[i].height(), 
                                 b.right());
         }

         if (onerow != null)
            onerow.add(c[i]);
      }
      
      if (b == null)
         b = new Rectangle(0,0,0,0);
      if (b2 == null)
         b2 = new Rectangle(0,0,0,0);
      
      headersHeight = b.top();
      verticalStep  = b.bottom() - b.top() + 1;
      
      if (isDynamic() && verticalStep == 0)
         verticalStep = 1;
      
      verticalStepBoxed  = b2.bottom() - b2.top() + 1;
   }

   /**
    * Relocates the current iteration's widgets for use with streamed output.
    * Relocation makes sure all widgets are placed as if they were of
    * iteration 1.
    */
   public void arrangeToStream()
   {
      if (!down)
         return;

      if (cursor <= 0)
         return;

      for (int j = 0; j < downBody[cursor].length; j++)
      {
         relocateComponent(downBody[0][j], 0, downBody[cursor][j]);
      }
   }

   /**
    * Relocates the current iteration's widgets for use with interactive
    * terminal after use with streamed output. 
    */
   public void arrangeToInteract()
   {
      if (!down)
         return;

      if (cursor <= 0)
         return;

      for (int j = 0; j < downBody[cursor].length; j++)
      {
         relocateComponent(downBody[0][j], startRow[cursor],
                           downBody[cursor][j]);
      }
   }

   /**
    * Rearranges widgets in the frame in preparation for the next drawing
    * of the frame in redirected mode. Does not apply to the interactive mode.
    * The need for this preparation is dictated by editors. Frames containing
    * no editors don't need it.
    * <p>
    * Editors can hold variable number of lines. If the widget's height does
    * not allow accommodation of all lines, the widget should be resized to get
    * the extra room. This leads to a cascade of relocations for all widgets
    * placed below the editor's rectangle. As the final step, the containers
    * holding the widgets, up to the frame, have to be resized accordingly.
    * <p>
    * During this rearranging process, the original dimensions and locations
    * of all affected widgets are saved for the subsequent restoration, which
    * is implemented in {@link #restoreEditors()} and is fast.
    */
   public void rearrangeEditors()
   {
      if (!hasEditors())
      {
         return;
      }

      int idx = 0;
      if (down)
      {
         idx = cursor;
      }

      if (idx < 0)
      {
         idx = 0;
      }

      // this is the frame height increment
      int totalIncr = 0;

      // this loop detects all editors, checks and adjusts their sizes
      for (int i = 0; i < downBody[idx].length; i++)
      {
         Widget comp = downBody[idx][i];

         if (comp instanceof Editor)
         {
            Editor ed = (Editor)comp;

            int oldh = ed.height();
            int newh = ed.fullHeight();

            // check to see if it has to be resized
            if (newh > oldh)
            {
               // to be resized
               int step = newh - oldh;
               int yloc = ed.location().y;

               // save the original height
               resizH.put(ed, Integer.valueOf(oldh));

               // resize the editor
               ed.setSize(ed.width(), ed.height() + step);

               // find all affected widgets; they have to be relocated
               for (int j = 0; j < downBody[idx].length; j++)
               {
                  Widget cmpn = downBody[idx][j];
                  if (cmpn.location().y >= yloc + oldh)
                  {
                     // this widget is affected
                     // save its original location if not done yet
                     if (!relocY.containsKey(cmpn))
                     {
                        relocY.put(cmpn, Integer.valueOf(cmpn.location().y));
                     }

                     // relocate the widget
                     Point loc = cmpn.location();
                     loc.y += step;
                     cmpn.setLocation(loc);

                     // invalidate the widget's rectangle
                     cmpn.repaint();
                  }
               }

               totalIncr += step;

               // invalidate the editor's rectangle
               ed.repaint();
            }
         }
      }

      // resize the frame
      origHeight = height();
      if (totalIncr > 0)
      {
         setSize(width(), origHeight + totalIncr);
      }

      // notify the redirected terminal about the ranges of editor lines
      ThinClient tc = ThinClient.getInstance();
      Stream stream = tc.getCurrentRedirection();
      RedirectedTerminal rt = ThinClient.redirectedTerminalForStream(stream);

      // this loop detects all editors and reports ranges of lines
      for (int i = 0; i < downBody[idx].length; i++)
      {
         Widget comp = downBody[idx][i];

         if (!(comp instanceof Editor))
         {
            continue;
         }

         Point loc = comp.screenLocation();
         int height = comp.height();

         rt.markBreakLines(loc.y, height);
      }
   }

   /**
    * Restores the dimensions and locations of all widgets as they were before
    * the call to {@link #rearrangeEditors()}.
    */
   public void restoreEditors()
   {
      if (!hasEditors())
      {
         return;
      }

      // shrink editors to their original heights
      Iterator<Editor> ite = resizH.keySet().iterator();
      while (ite.hasNext())
      {
         Editor ed = ite.next();
         Integer origH = resizH.get(ed);
         ed.setSize(ed.width(), origH.intValue());
      }

      // relocate all affected widgets back to their original locations
      Iterator<Widget> itc = relocY.keySet().iterator();
      while (itc.hasNext())
      {
         Widget co = itc.next();
         Point loc = co.location();
         loc.y = relocY.get(co).intValue();
         co.setLocation(loc);
      }

      // restore frame height
      setSize(width(), origHeight);
      origHeight = -1;

      // clear maps
      resizH.clear();
      relocY.clear();
   }

   /**
    * Reset forced refresh flag.
    */
   public void resetForcedRefresh()
   {
      forcedRefresh = false;
   }

   /**
    * Get current state of the forced refresh flag.
    * 
    * @return  the forced refresh flag.
    */
   public boolean isForcedRefresh()
   {
      return forcedRefresh;
   }

   /**
    * Finish pending cursor wrap.
    *
    * @param   normalExit
    *          Type of the exit, <code>true</code> means normal exit,
    *          <code>false</code> means exit with condition.
    * @param   doView
    *          Display saved iteration if any.
    */
   private void finishCursorWrap(boolean normalExit, boolean doView)
   {
      // repaint only if pause didn't trigger a condition
      if (normalExit)
      {
         clearWorker(true);
         repaint();
         cursor = pendingCursorPos;
         pendingCursorWrap = false;

         if (doView && spareIteration != null)
            spareIteration.view();

         spareIteration = null;
      }
      else if (!isRedirected())
      {
         forcedRefresh = true;
      }
   }

   /**
    * Create a copy of the given component shifted down at specified amount of
    * rows.
    * 
    * @param   component
    *          Component to make a copy of.
    *          
    * @return  A copy of the source component.
    */
   private Widget copyComponent(Widget component)
   {
      ComponentConfig src = component.config();
      Widget comp = UiUtils.reconstructComponent(src);
      
      comp.setParent(contentPane);
      
      if (comp instanceof RadioSet)
         ((RadioSet) comp).reset();
      
      comp.<WidgetConfig>config().setState(ScreenBuffer.UNINITIALIZED);
      
      return comp;
   }
   
   /**
    * Recalculate coordinates of the down frame body components.
    */
   private void readjustDownBody()
   {
      if (!down)
         return;

      int row = 0;
      startRow[0] = 0;
      
      lastRow = -1;
      
      for (int i = 1; i < downBody.length; i++)
      {
         row += underlinedRow[i - 1] ? 1 : verticalStep;
         startRow[i] = row;
         
         for (int j = 0; j < downBody[0].length; j++)
         {
            if (relocateComponent(downBody[0][j], row, downBody[i][j]) &&
                lastRow == -1)
            {
               lastRow = i - 1;
            }
         }
      }
      if (lastRow < 0)
         lastRow = downBody.length - 1;
   }

   /**
    * Get index of the component in the row.
    * 
    * @param   comp
    *          Component which is used as a reference.
    * 
    * @return  0-based index of the component in the row or -1 if component is
    *          not found.
    */
   private int getFieldIndex(Widget comp)
   {
      WidgetConfig cfg = comp.config();
      
      for (int i = 0; i < downBody[0].length; i++)
      {
         WidgetConfig tmp = downBody[0][i].config();
         if (tmp.getID() == cfg.getID())
            return i;
      }
      
      return -1;
   }

   /**
    * Relocate component to the given position using source component as a
    * reference.
    * 
    * @param   source
    *          Reference component.
    * @param   step
    *          Vertical step relative to source component.
    * @param   dest
    *          Destination component.
    * 
    * @return  <code>true</code> if component does not fit of the scroll pane
    *          and <code>false</code> otherwise.
    */
   protected boolean relocateComponent(Widget source, int step, Widget dest)
   {
      dest.setLocation(source.location().offset(0, step));
      
      if (dest instanceof LabeledWidget)
      {
         LabeledWidget lw = (LabeledWidget) dest;
         LabeledWidget lwsrc = (LabeledWidget) source;
         
         if (lwsrc.getLabelInstance() != null)
         {
            Label label = screen().getFactory().createLabel(lwsrc.getLabelInstance());
            label.setParent(contentPane);
            label.setLocation(label.location().offset(0, step));
            lw.setLabelInstance(label);
         }
         
         // label remains visible only if side labels are enabled
         if (!isSideLabels())
            lw.setNoLabels(true);
      }
      
      return (dest.location().y + dest.height() > frameScroll.height());
   }
   
   /**
    * Forces drawing of an empty line in redirected mode.
    */
   protected void forceNewLine()
   {
      screen().at(0, 0).append(' ', Color.DEFAULT).sync();
   }
   
   /**
    * Forces drawing of a box character for a number of consecutive lines
    * in redirected mode.
    *
    * @param    start
    *           line number in screen coordinates where the output starts
    * @param    lines
    *           number of lines to output to
    */
   protected void forceBox(int start, int lines)
   {
      screen().box(new Point(0, start),
                   new Dimension(1, lines),
                   Color.DEFAULT);
   }

   /**
    * Forces drawing of skips in redirected mode.
    * 
    * @param    where
    *           Vertical position.
    */
   protected void forceSkips(int where)
   {
      screen().at(0, where);
   }

   /**
    * Move the cursor to the component in the nearest row above this widget.
    * <p>
    * See {@link FrameFocusTransferManager#moveAboveWidget(Widget, int)} for
    * more details.
    *
    * @param   widget
    *          widget to check
    *
    * @param   validCursorPos
    *          cursor offset within the widget
    */
   public void moveAboveWidget(Widget widget, int validCursorPos)
   {
      focusTransferManager.moveAboveWidget(widget, validCursorPos);
   }

   /**
    * Move the cursor to the component in the nearest row below this widget.
    * <p>
    * See {@link FrameFocusTransferManager#moveBelowWidget(Widget, int)} for
    * more details.
    *
    * @param   widget
    *          widget to check
    *
    * @param   validCursorPos
    *          cursor offset within the widget
    */
   public void moveBelowWidget(Widget widget, int validCursorPos)
   {
      focusTransferManager.moveBelowWidget(widget, validCursorPos);
   }
   
   /**
    * Check if the specified widget is the left-most component in the current
    * screen row (not the same as the current record which may wrap).
    * <p>
    * See {@link FrameFocusTransferManager#isLeftMostWidget(Widget)} for
    * more details.
    *
    * @param   widget
    *          widget to check
    * 
    * @return  <code>true</code> if this component is the left-most component
    *          in the row.
    */
   public boolean isLeftMostWidget(Widget widget)
   {
      return focusTransferManager.isLeftMostWidget(widget);
   }

   /**
    * Check if the specified widget is the right-most component in the current
    * screen row (not the same as the current record which may wrap).
    * <p>
    * See {@link FrameFocusTransferManager#isRightMostWidget(Widget)} for
    * more details.
    * 
    * @param   widget
    *          widget to check
    * 
    * @return  <code>true</code> if this component is the right-most component
    *          in the row.
    */
   public boolean isRightMostWidget(Widget widget)
   {
      return focusTransferManager.isRightMostWidget(widget);
   }

   /**
    * Initialize CHOOSE processing mode.
    * 
    * @param   list
    *          List of components.
    * @param   data
    *          Choose data.
    * @param   focusOn
    *          ID of the widget to start from.
    *           
    * @return  ID of the selected widget or -1.
    */
   public Integer enterChoose(Widget[] list, ChooseData data, Integer focusOn)
   {
      return chooseHandler.enterChoose(list, data, focusId);
   }

   /**
    * Obtain current CHOOSE mode data.
    *
    * @return  Current instance.
    */
   public ChooseData getChooseData()
   {
      return chooseHandler.getChooseData();
   }

   /**
    * Stop CHOOSE mode.
    * 
    * @return  ChooseData instance as it is left after processing.
    */
   public ChooseData leaveChoose()
   {
      return chooseHandler.leaveChoose();
   }

   /**
    * Return the help text for the selected component using the CHOOSE index.
    *
    * @return   help text or <code>null</code> if not in CHOOSE
    */
   public String getChooseHelp()
   {
      return chooseHandler.getChooseHelp();
   }

   /**
    * Check to see if a widget is part of an ongoing CHOOSE.
    * 
    * @param   widget 
    *          Widget to check.
    * @return  <code>true</code> if the widget is part of a CHOOSE
    */
   public boolean isInChoose(Widget widget)
   {
      return chooseHandler.isInChoose(widget);
   }

   /**
    * Get current state of the choose activation flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isChooseActive()
   {
      return chooseHandler.isChooseActive();
   }

   /**
    * Positions the CHOOSE at the specified widget if possible and returns
    * the indication of success or failure.
    * If the specified widget is not part of the active choose, the choose
    * repositions to nowhere and the highlight gets removed from the frame.
    * 
    * @param   widget 
    *          Widget to set current CHOOSE position at.
    * @return  <code>true</code> if the widget is part of an active CHOOSE
    */
   public boolean setChooseAt(Widget widget)
   {
      return chooseHandler.setChooseAt(widget);
   }

   /**
    * Looks up the very first enabled widget in specified frames.
    * 
    * @return  the first enabled widget or <code>null</code>
    */
   public Widget getFirstEnabledWidget()
   {
      // find the first enabled widget of this frame
      Widget[] widgets = getContentPane().widgets();
      Widget comp = null;
   
      for (int i = 0; i < widgets.length; i++)
      {
         comp = getField(widgets[i]);
         
         if (UiUtils.hasConfig(comp) && comp.focusTraversable())
         {
            return comp;
         }
      }
   
      // no enabled widgets in all frames
      return null; 
   }

   /**
    * Enable widgets specified in provided list.
    * 
    * @param   affw
    *          Storage for affected widgets.
    * @param   focus
    *          Focused widget.
    * @param   widgetId
    *          List of widget IDs to enable.
    * @param   enabled
    *          New enabled state for the affected widgets.
    *          
    * @return  Affected widget IDs as set.
    */
   public Set<Integer> enabledWidgetsByList(Widget[] affw, Widget focus,
                                     int[] widgetId, boolean enabled)
   {
      Widget comp = null;
      Set<Integer> ids = new HashSet<Integer>();
      FrameConfig cfg = config();
      ThinClient thinClient = ThinClient.getInstance();
      
      for (int i = 0; i < widgetId.length; i++)
      {
         comp = thinClient.getWidget(widgetId[i]);
         
         comp = getField(comp);
      
         if (UiUtils.isFocusable(comp))
         {
            comp.setEnabled(enabled);
            if (comp == focus && !enabled)
            {
               // save the next enabled widget, which may be used
               // by APPLY
               Widget enabledWidget = getNextEnabledWidget(comp, 
                                                           FocusManager.FOCUS_WRAP);
               thinClient.setPendingFocusApply(enabledWidget);
            }
         }
      
         if (enabled)
         {
            if (!comp.isVisible() && (isVisible() || !hidden()))
            {
               thinClient.collectFlag(comp, true, true);
               comp.setVisible(true);
            }
      
            thinClient.collectFlag(comp, false, false);
            comp.setHidden(false);
         }
      
         // maintain tab order
         if (enabled && comp.focusTraversable()  && !cfg.isKeepTabOrder())
         {
            getContentPane().moveToTop(comp);
         }
      
         // copy the affected widgets into the list
         affw[i] = comp;
         ids.add(widgetId[i]);
      }
      
      return ids;
   }

   /**
    * Looks up the first focusable widget of the frame.
    * 
    * @param   frame
    *          frame to query
    *          
    * @return  widget ID of the first focusable widget or <code>-1</code>
    *          if no focusable widget is found
    */
   public static int getFirstFocusableId(Frame frame)
   {
      // get list of all widgets in this frame
      Widget[] list  = frame.getContentPane().widgets();
      
      // scan the list
      for (int i = 0; i < list.length; i++)
      {
         if (list[i].config() == null)
         {
            continue;
         }
   
         if (!list[i].isEnabled())
         {
            continue;
         }
   
         if (!list[i].focusTraversable())
         {
            continue;
         }
   
         return list[i].<WidgetConfig>config().getID();
      }
   
      return -1;
   }

   /**
    * Looks up the next enabled and visible widget in the tab order of the 
    * specified frame.
    *
    * @param  widget
    *         widget to start search after 
    * @param  wrap
    *         if <code>true</code> the search wraps to the first widget
    *         of the frame when the last widget fails
    *
    * @return the next enabled and visible widget or <code>null</code>
    */
   public Widget getNextEnabledWidget(Widget widget, boolean wrap)
   {
      // get all widgets of this frame
      Widget[] widgets = getContentPane().widgets();
   
      // find the reference point
      int nref = 0;
      for (nref = 0; nref < widgets.length; nref++)
      {
         if (widgets[nref] == widget)
         {
            break;
         }
      }
   
      // if no reference point, assume no result
      if (nref == widgets.length)
      {
         return null;
      }
   
      // find the next enabled widget with no wrapping first
      Widget comp = null;
   
      for (int i = nref + 1; i < widgets.length; i++)
      {
         comp = getField(widgets[i]);
         if (UiUtils.hasConfig(comp) && comp.focusTraversable())
         {
            return comp;
         }
      }
   
      // if wrapping is required, check the first section of the list
      if (wrap && nref > 0)
      {
         for (int i = 0; i < nref; i++)
         {
            comp = getField(widgets[i]);
            if (UiUtils.hasConfig(comp) && comp.focusTraversable())
            {
               return comp;
            }
         }
      }
   
      // no enabled widgets in this frame
      return null; 
   }

   /**
    * Enable/disable all widgets in frame.
    * 
    * @param   focus
    *          Focused widget.
    * @param   enabled
    *          New enabled state.
    *          
    * @return  number of affected widgets.
    */
   public int enableAllWidgets(Widget focus, boolean enabled)
   {
      // enumerate all widgets in the frame and enable/disable all
      // field level widgets
      int naffw = 0;
      Widget comp = null;
      Widget[] list = getContentPane().widgets();
      ThinClient thinClient = ThinClient.getInstance();
      
      for (int i = 0; i < list.length; i++)
      {
         comp = getField(list[i]);
         
         if (!UiUtils.isFocusable(comp))
            continue;
         
         comp.setEnabled(enabled);
         if (comp == focus && !enabled)
         {
            // save the next enabled widget, which may be used
            // by APPLY
            Widget widget = getNextEnabledWidget(comp, FocusManager.FOCUS_WRAP);
            thinClient.setPendingFocusApply(widget);
         }
         naffw++;
    
         if (enabled)
         {
            if (!comp.isVisible())
            {
               thinClient.collectFlag(comp, true, true);
               comp.setVisible(true);
            }
    
            thinClient.collectFlag(comp, false, false);
            comp.setHidden(false);
         }
      }
      return naffw;
   }

   /**
    * Copy affected widgets.
    * 
    * @param   naffw
    *          Number of widgets.
    *          
    * @return  Arrays of affected widgets.
    */
   public Widget[] copyAffectedWidgets(int naffw)
   {
      Widget[] list = getContentPane().widgets();
      Widget[] affw = new Widget[naffw];
      Widget comp;
      
      if (naffw > 0)
      {
         affw = new Widget[naffw];
         int j = 0;
         for (int i = 0; i < list.length; i++)
         {
            comp = getField(list[i]);
            
            if (UiUtils.isFocusable(comp))
            {
               affw[j] = comp;
               j++;
            }
         }
      }
      return affw;
   }

   /**
    * Switch empty mode for labels which fit into specified rectangle.
    * 
    * @param   rect
    *          Input rectangle. 
    */
   public void switchLabelsEmptyMode(Rectangle rect)
   {
      Widget[] cmp = getContentPane().widgets();
      
      for (Widget component : cmp)
      {
         if (component instanceof Label && 
             rect.intersects(new Rectangle(component.screenLocation(), 
                                           component.dimension())))
         {
            ((Label) component).setEmptyModeOff();
            ((Label) component).setEmptyModeOn();
         }
            
      }
   }
}
