/*
** Module   : AsyncRequest.java
** Abstract : defines the async request resource, created during async procedure calls. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130813 Created initial version.
** 002 CA  20130927 Added LegacyResource annotation, to determine resource type.
*/

package com.goldencode.p2j.util;

/**
 * Define the async request resource, created during async procedure calls. Adds support for these
 * attributes and methods:
 * <table>
 * <tr>
 *    <td>CANCELLED</td>
 *    <td>{@link #isCancelled()}</td>
 * </tr>
 * <tr>
 *    <td>COMPLETE</td>
 *    <td>{@link #isComplete()}</td>
 * </tr>
 * <tr>
 *    <td>EVENT-PROCEDURE</td>
 *    <td>{@link #getEventProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>EVENT-PROCEDURE-CONTEXT</td>
 *    <td>{@link #getEventProcedureContext()}</td>
 * </tr>
 * <tr>
 *    <td>PERSISTENT-PROCEDURE</td>
 *    <td>{@link #getPersistentProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>PROCEDURE-NAME</td>
 *    <td>{@link #getProcedureName()}</td>
 * </tr>
 * <tr>
 *    <td>QUIT</td>
 *    <td>{@link #isQuit()}</td>
 * </tr>
 * <tr>
 *    <td>STOP</td>
 *    <td>{@link #isStop()}</td>
 * </tr>
 * <tr>
 *    <td>SERVER</td>
 *    <td>{@link #getServerHandle()}</td>
 * </tr>
 * </table>
 */
@LegacyResource(resource = LegacyResource.ASYNC_REQUEST)
public interface AsyncRequest
extends CommonHandleChain,
        Remotable,
        Errorable
{
   /**
    * Get the <code>CANCELLED</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public logical isCancelled();

   /**
    * Get the <code>COMPLETE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public logical isComplete();

   /**
    * Get the <code>EVENT-PROCEDURE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public character getEventProcedure();

   /**
    * Get the <code>EVENT-PROCEDURE-CONTEXT</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public handle getEventProcedureContext();

   /**
    * Get the <code>PERSISTENT-PROCEDURE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public handle getPersistentProcedure();

   /**
    * Get the <code>PROCEDURE-NAME</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public character getProcedureName();

   /**
    * Get the <code>QUIT</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public logical isQuit();

   /**
    * Get the <code>STOP</code> attribute of this async request.
    * 
    * @return   See above.
    */
   public logical isStop();
}
